from __future__ import annotations

from dataclasses import field
from typing import Optional

import click

from chik.cmds.cmd_classes import ChikCliContext, chik_command, option
from chik.cmds.cmd_helpers import NeedsWalletRPC
from chik.cmds.param_types import (
    AddressParamType,
    Bytes32ParamType,
    CliAddress,
    TransactionFeeParamType,
)
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.errors import CliRpcConnectionError
from chik.util.ints import uint64


@click.group("plotnft", help="Manage your plot NFTs")
@click.pass_context
def plotnft_cmd(ctx: click.Context) -> None:
    pass


@chik_command(
    group=plotnft_cmd,
    name="show",
    short_help="Show plotnft information",
    help="Show plotnft information",
)
class ShowPlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    context: ChikCliContext = field(default_factory=ChikCliContext)
    id: Optional[int] = option(
        "-i", "--id", help="ID of the wallet to use", default=None, show_default=True, required=False
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import show

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await show(
                wallet_info=wallet_info,
                root_path=self.context.root_path,
                wallet_id_passed_in=self.id,
            )


@chik_command(
    group=plotnft_cmd,
    name="get_login_link",
    short_help="Create a login link for a pool",
    help="Create a login link for a pool. The farmer must be running. Use 'plotnft show' to get the launcher id.",
)
class GetLoginLinkCMD:
    context: ChikCliContext = field(default_factory=ChikCliContext)
    launcher_id: bytes32 = option(
        "-l", "--launcher_id", help="Launcher ID of the plotnft", type=Bytes32ParamType(), required=True
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import get_login_link

        await get_login_link(self.launcher_id, root_path=self.context.root_path)


# Functions with this mark in this file are not being ported to @tx_out_cmd due to lack of observer key support
# They will therefore not work with observer-only functionality
# NOTE: tx_endpoint  (This creates wallet transactions and should be parametrized by relevant options)
@chik_command(
    group=plotnft_cmd,
    name="create",
    short_help="Create a plot NFT",
    help="Create a plot NFT.",
)
class CreatePlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    pool_url: Optional[str] = option("-u", "--pool-url", help="HTTPS host:port of the pool to join", required=False)
    state: str = option(
        "-s",
        "--state",
        help="Initial state of Plot NFT: local or pool",
        required=True,
        type=click.Choice(["local", "pool"], case_sensitive=False),
    )
    fee: uint64 = option(
        "-m",
        "--fee",
        help="Set the fees per transaction, in XCK. Fee is used TWICE: once to create the singleton, once for init.",
        type=TransactionFeeParamType(),
        default="0",
        show_default=True,
        required=True,
    )
    dont_prompt: bool = option("-y", "--yes", help="No prompts", is_flag=True)

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import create

        if self.pool_url is not None and self.state == "local":
            raise CliRpcConnectionError(f"A pool url [{self.pool_url}] is not allowed with 'local' state")

        if self.pool_url in {None, ""} and self.state == "pool":
            raise CliRpcConnectionError("A pool url argument (-u/--pool-url) is required with 'pool' state")

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await create(
                wallet_info=wallet_info,
                pool_url=self.pool_url,
                state="FARMING_TO_POOL" if self.state == "pool" else "SELF_POOLING",
                fee=self.fee,
                prompt=not self.dont_prompt,
            )


# NOTE: tx_endpoint
@chik_command(
    group=plotnft_cmd,
    name="join",
    short_help="Join a plot NFT to a Pool",
    help="Join a plot NFT to a Pool.",
)
class JoinPlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    pool_url: str = option("-u", "--pool-url", help="HTTPS host:port of the pool to join", required=True)
    fee: uint64 = option(
        "-m",
        "--fee",
        help="Set the fees per transaction, in XCK. Fee is used TWICE: once to create the singleton, once for init.",
        type=TransactionFeeParamType(),
        default="0",
        show_default=True,
        required=True,
    )
    dont_prompt: bool = option("-y", "--yes", help="No prompts", is_flag=True)
    id: Optional[int] = option(
        "-i", "--id", help="ID of the wallet to use", default=None, show_default=True, required=False
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import join_pool

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await join_pool(
                wallet_info=wallet_info,
                pool_url=self.pool_url,
                fee=self.fee,
                wallet_id=self.id,
                prompt=not self.dont_prompt,
            )


# NOTE: tx_endpoint
@chik_command(
    group=plotnft_cmd,
    name="leave",
    short_help="Leave a pool and return to self-farming",
    help="Leave a pool and return to self-farming.",
)
class LeavePlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    dont_prompt: bool = option("-y", "--yes", help="No prompts", is_flag=True)
    fee: uint64 = option(
        "-m",
        "--fee",
        help="Set the fees per transaction, in XCK. Fee is used TWICE: once to create the singleton, once for init.",
        type=TransactionFeeParamType(),
        default="0",
        show_default=True,
        required=True,
    )
    id: Optional[int] = option(
        "-i", "--id", help="ID of the wallet to use", default=None, show_default=True, required=False
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import self_pool

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await self_pool(
                wallet_info=wallet_info,
                fee=self.fee,
                wallet_id=self.id,
                prompt=not self.dont_prompt,
            )


@chik_command(
    group=plotnft_cmd,
    name="inspect",
    short_help="Get Detailed plotnft information as JSON",
    help="Get Detailed plotnft information as JSON",
)
class InspectPlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    id: Optional[int] = option(
        "-i", "--id", help="ID of the wallet to use", default=None, show_default=True, required=False
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import inspect_cmd

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await inspect_cmd(wallet_info=wallet_info, wallet_id=self.id)


# NOTE: tx_endpoint
@chik_command(
    group=plotnft_cmd,
    name="claim",
    short_help="Claim rewards from a plot NFT",
    help="Claim rewards from a plot NFT",
)
class ClaimPlotNFTCMD:
    rpc_info: NeedsWalletRPC  # provides wallet-rpc-port and fingerprint options
    id: Optional[int] = option(
        "-i", "--id", help="ID of the wallet to use", default=None, show_default=True, required=False
    )
    fee: uint64 = option(
        "-m",
        "--fee",
        help="Set the fees per transaction, in XCK. Fee is used TWICE: once to create the singleton, once for init.",
        type=TransactionFeeParamType(),
        default="0",
        show_default=True,
        required=True,
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import claim_cmd

        async with self.rpc_info.wallet_rpc() as wallet_info:
            await claim_cmd(
                wallet_info=wallet_info,
                fee=self.fee,
                wallet_id=self.id,
            )


@chik_command(
    group=plotnft_cmd,
    name="change_payout_instructions",
    short_help="Change the payout instructions for a pool.",
    help="Change the payout instructions for a pool. Use 'plotnft show' to get the launcher id.",
)
class ChangePayoutInstructionsPlotNFTCMD:
    context: ChikCliContext = field(default_factory=ChikCliContext)
    launcher_id: bytes32 = option(
        "-l", "--launcher_id", help="Launcher ID of the plotnft", type=Bytes32ParamType(), required=True
    )
    address: CliAddress = option(
        "-a", "--address", help="New address for payout instructions", type=AddressParamType(), required=True
    )

    async def run(self) -> None:
        from chik.cmds.plotnft_funcs import change_payout_instructions

        await change_payout_instructions(self.launcher_id, self.address, root_path=self.context.root_path)



from __future__ import annotations

import pytest
from klvm.EvalError import EvalError
from klvm.operators import KEYWORD_TO_ATOM
from klvm_tools.binutils import assemble, disassemble

from chik.types.blockchain_format.program import Program


def test_at():
    p = Program.to([10, 20, 30, [15, 17], 40, 50])

    assert p.first() == p.at("f")
    assert Program.to(10) == p.at("f")

    assert p.rest() == p.at("r")
    assert Program.to([20, 30, [15, 17], 40, 50]) == p.at("r")

    assert p.rest().rest().rest().first().rest().first() == p.at("rrrfrf")
    assert Program.to(17) == p.at("rrrfrf")

    with pytest.raises(ValueError):
        p.at("q")

    with pytest.raises(EvalError):
        p.at("ff")


def test_replace():
    p1 = Program.to([100, 200, 300])
    assert p1.replace(f=105) == Program.to([105, 200, 300])
    assert p1.replace(rrf=[301, 302]) == Program.to([100, 200, [301, 302]])
    assert p1.replace(f=105, rrf=[301, 302]) == Program.to([105, 200, [301, 302]])
    assert p1.replace(f=100, r=200) == Program.to((100, 200))


def test_replace_conflicts():
    p1 = Program.to([100, 200, 300])
    with pytest.raises(ValueError):
        p1.replace(rr=105, rrf=200)


def test_replace_conflicting_paths():
    p1 = Program.to([100, 200, 300])
    with pytest.raises(ValueError):
        p1.replace(ff=105)


def test_replace_bad_path():
    p1 = Program.to([100, 200, 300])
    with pytest.raises(ValueError):
        p1.replace(q=105)
    with pytest.raises(ValueError):
        p1.replace(rq=105)


def check_idempotency(f, *args):
    prg = Program.to(f)
    curried = prg.curry(*args)

    r = disassemble(curried)
    f_0, args_0 = curried.uncurry()

    assert disassemble(f_0) == disassemble(f)
    assert disassemble(args_0) == disassemble(Program.to(list(args)))
    return r


def test_curry_uncurry():
    PLUS = KEYWORD_TO_ATOM["+"][0]
    f = assemble("(+ 2 5)")
    actual_disassembly = check_idempotency(f, 200, 30)
    assert actual_disassembly == f"(a (q {PLUS} 2 5) (c (q . 200) (c (q . 30) 1)))"

    f = assemble("(+ 2 5)")
    args = assemble("(+ (q . 50) (q . 60))")
    # passing "args" here wraps the arguments in a list
    actual_disassembly = check_idempotency(f, args)
    assert actual_disassembly == f"(a (q {PLUS} 2 5) (c (q {PLUS} (q . 50) (q . 60)) 1))"


def test_uncurry_not_curried():
    # this function has not been curried
    plus = Program.to(assemble("(+ 2 5)"))
    assert plus.uncurry() == (plus, Program.to(0))


def test_uncurry():
    # this is a positive test
    plus = Program.to(assemble("(2 (q . (+ 2 5)) (c (q . 1) 1))"))
    assert plus.uncurry() == (Program.to(assemble("(+ 2 5)")), Program.to([1]))


def test_uncurry_top_level_garbage():
    # there's garbage at the end of the top-level list
    plus = Program.to(assemble("(2 (q . 1) (c (q . 1) (q . 1)) (q . 0x1337))"))
    assert plus.uncurry() == (plus, Program.to(0))


def test_uncurry_not_pair():
    # the second item in the list is expected to be a pair, with a quote
    plus = Program.to(assemble("(2 1 (c (q . 1) (q . 1)))"))
    assert plus.uncurry() == (plus, Program.to(0))


def test_uncurry_args_garbage():
    # there's garbage at the end of the args list
    plus = Program.to(assemble("(2 (q . 1) (c (q . 1) (q . 1) (q . 0x1337)))"))
    assert plus.uncurry() == (plus, Program.to(0))


def test_run() -> None:
    div = Program.to(assemble("(/ 2 5)"))
    ret = div.run([10, 5])
    assert ret.atom == bytes([2])

    ret = div.run([10, -5])
    assert ret.atom == bytes([0xFE])

    # run()
    cost, ret = div.run_with_cost(100000, [10, -5], 0)
    assert cost == 1107
    print(ret)
    assert ret.atom == bytes([0xFE])

    cost, ret = div.run_with_cost(100000, [10, -5], 0)
    assert cost == 1107
    print(ret)
    assert ret.atom == bytes([0xFE])

    # run_with_flags()
    cost, ret = div.run_with_flags(100000, 0, [10, -5])
    assert cost == 1107
    print(ret)
    assert ret.atom == bytes([0xFE])

    cost, ret = div.run_with_flags(100000, 0, [10, -5])
    assert cost == 1107
    print(ret)
    assert ret.atom == bytes([0xFE])

    # run_with_cost()
    ret = div.run([10, -5], 100000, 0)
    print(ret)
    assert ret.atom == bytes([0xFE])

    ret = div.run([10, -5], 100000, 0)
    print(ret)
    assert ret.atom == bytes([0xFE])



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional, TypeVar

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint8, uint32, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.conditions import ConditionValidTimes
from chik.wallet.trading.offer import Offer
from chik.wallet.trading.trade_status import TradeStatus

_T_TradeRecord = TypeVar("_T_TradeRecord", bound="TradeRecordOld")


@streamable
@dataclass(frozen=True)
class TradeRecordOld(Streamable):
    """
    Used for storing transaction data and status in wallets.
    """

    confirmed_at_index: uint32
    accepted_at_time: Optional[uint64]
    created_at_time: uint64
    is_my_offer: bool
    sent: uint32
    offer: bytes
    taken_offer: Optional[bytes]
    coins_of_interest: list[Coin]
    trade_id: bytes32
    status: uint32  # TradeStatus, enum not streamable
    sent_to: list[tuple[str, uint8, Optional[str]]]  # MempoolSubmissionStatus.status enum not streamable

    def to_json_dict_convenience(self) -> dict[str, Any]:
        formatted = self.to_json_dict()
        formatted["status"] = TradeStatus(self.status).name
        offer_to_summarize: bytes = self.offer if self.taken_offer is None else self.taken_offer
        offer = Offer.from_bytes(offer_to_summarize)
        offered, requested, infos, _ = offer.summary()
        formatted["summary"] = {
            "offered": offered,
            "requested": requested,
            "infos": infos,
            "fees": offer.fees(),
        }
        formatted["pending"] = offer.get_pending_amounts()
        del formatted["offer"]
        return formatted

    @classmethod
    def from_json_dict_convenience(
        cls: type[_T_TradeRecord], record: dict[str, Any], offer: str = ""
    ) -> _T_TradeRecord:
        new_record = record.copy()
        new_record["status"] = TradeStatus[record["status"]].value
        del new_record["summary"]
        del new_record["pending"]
        new_record["offer"] = offer
        return cls.from_json_dict(new_record)


@streamable
@dataclass(frozen=True)
class TradeRecord(TradeRecordOld):
    valid_times: ConditionValidTimes



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint32, uint64
from chik.util.streamable import Streamable, streamable
from chik.wallet.lineage_proof import LineageProof


@streamable
@dataclass(frozen=True)
class ProposalInfo(Streamable):
    proposal_id: bytes32  # this is launcher_id
    inner_puzzle: Program
    amount_voted: uint64
    yes_votes: uint64
    current_coin: Coin
    current_innerpuz: Optional[Program]
    timer_coin: Optional[Coin]  # if this is None then the proposal has finished
    singleton_block_height: uint32  # Block height that current proposal singleton coin was created in
    passed: bool
    closed: bool


@streamable
@dataclass(frozen=True)
class DAOInfo(Streamable):
    treasury_id: bytes32
    cat_wallet_id: uint32
    dao_cat_wallet_id: uint32
    proposals_list: list[ProposalInfo]
    parent_info: list[tuple[bytes32, Optional[LineageProof]]]  # {coin.name(): LineageProof}
    current_treasury_coin: Optional[Coin]
    current_treasury_innerpuz: Optional[Program]
    singleton_block_height: uint32  # the block height that the current treasury singleton was created in
    filter_below_vote_amount: uint64  # we ignore proposals with fewer votes than this - defaults to 1
    assets: list[Optional[bytes32]]
    current_height: uint64


@streamable
@dataclass(frozen=True)
class DAORules(Streamable):
    proposal_timelock: uint64
    soft_close_length: uint64
    attendance_required: uint64
    pass_percentage: uint64
    self_destruct_length: uint64
    oracle_spend_delay: uint64
    proposal_minimum_amount: uint64


class ProposalType(Enum):
    SPEND = "s"
    UPDATE = "u"
    MINT = "m"



from __future__ import annotations

import argparse
import contextlib
import copy
import logging
import os
import shutil
import sys
import tempfile
import time
import traceback
from collections.abc import Iterator
from pathlib import Path
from typing import Any, Callable, Optional, Union, cast

import importlib_resources
import yaml
from typing_extensions import Literal

from chik.server.outbound_message import NodeType
from chik.types.peer_info import UnresolvedPeerInfo
from chik.util.lock import Lockfile

log = logging.getLogger(__name__)


def initial_config_file(filename: Union[str, Path]) -> str:
    initial_config_path = importlib_resources.files(__name__.rpartition(".")[0]).joinpath(f"initial-{filename}")
    contents: str = initial_config_path.read_text(encoding="utf-8")
    return contents


def create_default_chik_config(root_path: Path, filenames: list[str] = ["config.yaml"]) -> None:
    for filename in filenames:
        default_config_file_data: str = initial_config_file(filename)
        path: Path = config_path_for_filename(root_path, filename)
        tmp_path: Path = path.with_suffix("." + str(os.getpid()))
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(tmp_path, "w") as f:
            f.write(default_config_file_data)
        try:
            os.replace(str(tmp_path), str(path))
        except PermissionError:
            shutil.move(str(tmp_path), str(path))


def config_path_for_filename(root_path: Path, filename: Union[str, Path]) -> Path:
    path_filename = Path(filename)
    if path_filename.is_absolute():
        return path_filename
    return root_path / "config" / filename


@contextlib.contextmanager
def lock_config(root_path: Path, filename: Union[str, Path]) -> Iterator[None]:
    # TODO: This is presently used in some tests to lock the saving of the
    #       configuration file without having loaded it right there.  This usage
    #       should probably be removed and this function made private.
    config_path = config_path_for_filename(root_path, filename)
    with Lockfile.create(config_path):
        yield


@contextlib.contextmanager
def lock_and_load_config(
    root_path: Path,
    filename: Union[str, Path],
    fill_missing_services: bool = False,
) -> Iterator[dict[str, Any]]:
    with lock_config(root_path=root_path, filename=filename):
        config = _load_config_maybe_locked(
            root_path=root_path,
            filename=filename,
            acquire_lock=False,
            fill_missing_services=fill_missing_services,
        )
        yield config


def save_config(root_path: Path, filename: Union[str, Path], config_data: Any) -> None:
    # This must be called under an acquired config lock
    path: Path = config_path_for_filename(root_path, filename)
    with tempfile.TemporaryDirectory(dir=path.parent) as tmp_dir:
        tmp_path: Path = Path(tmp_dir) / Path(filename)
        with open(tmp_path, "w") as f:
            yaml.safe_dump(config_data, f)
        try:
            os.replace(str(tmp_path), path)
        except PermissionError:
            shutil.move(str(tmp_path), str(path))


def load_config(
    root_path: Path,
    filename: Union[str, Path],
    sub_config: Optional[str] = None,
    exit_on_error: bool = True,
    fill_missing_services: bool = False,
) -> dict[str, Any]:
    return _load_config_maybe_locked(
        root_path=root_path,
        filename=filename,
        sub_config=sub_config,
        exit_on_error=exit_on_error,
        acquire_lock=True,
        fill_missing_services=fill_missing_services,
    )


def _load_config_maybe_locked(
    root_path: Path,
    filename: Union[str, Path],
    sub_config: Optional[str] = None,
    exit_on_error: bool = True,
    acquire_lock: bool = True,
    fill_missing_services: bool = False,
) -> dict[str, Any]:
    # This must be called under an acquired config lock, or acquire_lock should be True

    path = config_path_for_filename(root_path, filename)

    if not path.is_file():
        if not exit_on_error:
            raise ValueError("Config not found")
        print(f"can't find {path}")
        print("** please run `chik init` to migrate or create new config files **")
        # TODO: fix this hack
        sys.exit(-1)
    # This loop should not be necessary due to the config lock, but it's kept here just in case
    for i in range(10):
        try:
            # at least we intend it to be this type
            r: dict[str, Any]
            with contextlib.ExitStack() as exit_stack:
                if acquire_lock:
                    exit_stack.enter_context(lock_config(root_path, filename))
                with open(path) as opened_config_file:
                    r = yaml.safe_load(opened_config_file)
            if r is None:
                log.error(f"yaml.safe_load returned None: {path}")
                time.sleep(i * 0.1)
                continue
            if fill_missing_services:
                r.update(load_defaults_for_missing_services(config=r, config_name=path.name))
            if sub_config is not None:
                r = cast(dict[str, Any], r.get(sub_config))
            return r
        except Exception as e:
            tb = traceback.format_exc()
            log.error(f"Error loading file: {tb} {e} Retrying {i}")
            time.sleep(i * 0.1)
    raise RuntimeError("Was not able to read config file successfully")


def load_config_cli(
    root_path: Path,
    filename: str,
    sub_config: Optional[str] = None,
    fill_missing_services: bool = False,
) -> dict[str, Any]:
    """
    Loads configuration from the specified filename, in the config directory,
    and then overrides any properties using the passed in command line arguments.
    Nested properties in the config file can be used in the command line with ".",
    for example --farmer_peer.host. Does not support lists.
    """
    config = load_config(root_path, filename, sub_config, fill_missing_services=fill_missing_services)

    flattened_props = flatten_properties(config)
    parser = argparse.ArgumentParser()

    for prop_name, value in flattened_props.items():
        if type(value) is list:
            continue
        prop_type: Callable = str2bool if type(value) is bool else type(value)  # type: ignore
        parser.add_argument(f"--{prop_name}", type=prop_type, dest=prop_name)

    for key, value in vars(parser.parse_args()).items():
        if value is not None:
            flattened_props[key] = value

    return unflatten_properties(flattened_props)


def flatten_properties(config: dict[str, Any]) -> dict[str, Any]:
    properties = {}
    for key, value in config.items():
        if type(value) is dict:
            for key_2, value_2 in flatten_properties(value).items():
                properties[key + "." + key_2] = value_2
        else:
            properties[key] = value
    return properties


def unflatten_properties(config: dict[str, Any]) -> dict[str, Any]:
    properties: dict[str, Any] = {}
    for key, value in config.items():
        if "." in key:
            add_property(properties, key, value)
        else:
            properties[key] = value
    return properties


def add_property(d: dict[str, Any], partial_key: str, value: Any) -> None:
    if "." not in partial_key:  # root of dict
        d[partial_key] = value
    else:
        key_1, key_2 = partial_key.split(".", maxsplit=1)
        if key_1 not in d:
            d[key_1] = {}
        if "." in key_2:
            add_property(d[key_1], key_2, value)
        else:
            d[key_1][key_2] = value


def str2bool(v: Union[str, bool]) -> bool:
    # Source from https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
    if isinstance(v, bool):
        return v
    if v.lower() in {"yes", "true", "True", "t", "y", "1"}:
        return True
    elif v.lower() in {"no", "false", "False", "f", "n", "0"}:
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def traverse_dict(d: dict[str, Any], key_path: str) -> Any:
    """
    Traverse nested dictionaries to find the element pointed-to by key_path.
    Key path components are separated by a ':' e.g.
      "root:child:a"
    """
    if type(d) is not dict:
        raise TypeError(f"unable to traverse into non-dict value with key path: {key_path}")

    # Extract one path component at a time
    components = key_path.split(":", maxsplit=1)
    if components is None or len(components) == 0:
        raise KeyError(f"invalid config key path: {key_path}")

    key = components[0]
    remaining_key_path = components[1] if len(components) > 1 else None

    val: Any = d.get(key, None)
    if val is not None:
        if remaining_key_path is not None:
            return traverse_dict(val, remaining_key_path)
        return val
    else:
        raise KeyError(f"value not found for key: {key}")


method_strings = Literal["default", "python_default", "fork", "forkserver", "spawn"]
method_values = Optional[Literal["fork", "forkserver", "spawn"]]
start_methods: dict[method_strings, method_values] = {
    "default": None,
    "python_default": None,
    "fork": "fork",
    "forkserver": "forkserver",
    "spawn": "spawn",
}


def process_config_start_method(
    config: dict[str, Any],
    log: logging.Logger,
) -> method_values:
    from_config: object = config.get("multiprocessing_start_method")

    choice: method_strings
    if from_config is None:
        # handle not only the key being missing, but also set to None
        choice = "default"
    elif from_config not in start_methods.keys():
        start_methods_string = ", ".join(option for option in start_methods.keys())
        log.warning(f"Configured start method {from_config!r} not available in: {start_methods_string}")
        choice = "default"
    else:
        # mypy doesn't realize that by the time we get here from_config must be one of
        # the keys in `start_methods` due to the above `not in` condition.
        choice = from_config  # type: ignore[assignment]

    processed_method = start_methods[choice]
    log.info(f"Selected multiprocessing start method: {choice}")

    return processed_method


def override_config(config: dict[str, Any], config_overrides: Optional[dict[str, Any]]) -> dict[str, Any]:
    new_config = copy.deepcopy(config)
    if config_overrides is None:
        return new_config
    for k, v in config_overrides.items():
        add_property(new_config, k, v)
    return new_config


def selected_network_address_prefix(config: dict[str, Any]) -> str:
    # we intend this to be a str at least
    address_prefix: str = config["network_overrides"]["config"][config["selected_network"]]["address_prefix"]
    return address_prefix


def load_defaults_for_missing_services(config: dict[str, Any], config_name: str) -> dict[str, Any]:
    services = ["data_layer"]
    missing_services = [service for service in services if service not in config]
    defaulted = {}
    if len(missing_services) > 0:
        marshalled_default_config: str = initial_config_file(config_name)

        unmarshalled_default_config = yaml.safe_load(marshalled_default_config)

        for service in missing_services:
            defaulted[service] = unmarshalled_default_config[service]

            if "logging" in defaulted[service]:
                to_be_referenced = config.get("logging")
                if to_be_referenced is not None:
                    defaulted[service]["logging"] = to_be_referenced

            if "selected_network" in defaulted[service]:
                to_be_referenced = config.get("selected_network")
                if to_be_referenced is not None:
                    # joining to hopefully force a new string of the same value
                    defaulted[service]["selected_network"] = "".join(to_be_referenced)

    return defaulted


PEER_INFO_MAPPING: dict[NodeType, str] = {
    NodeType.FULL_NODE: "full_node_peer",
    NodeType.FARMER: "farmer_peer",
}


def get_unresolved_peer_infos(service_config: dict[str, Any], peer_type: NodeType) -> set[UnresolvedPeerInfo]:
    peer_info_key = PEER_INFO_MAPPING[peer_type]
    peer_infos: list[dict[str, Any]] = service_config.get(f"{peer_info_key}s", [])
    peer_info: Optional[dict[str, Any]] = service_config.get(peer_info_key)
    if peer_info is not None:
        peer_infos.append(peer_info)

    return {UnresolvedPeerInfo(host=peer["host"], port=peer["port"]) for peer in peer_infos}


def set_peer_info(
    service_config: dict[str, Any],
    peer_type: NodeType,
    peer_host: Optional[str] = None,
    peer_port: Optional[int] = None,
) -> None:
    peer_info_key = PEER_INFO_MAPPING[peer_type]
    if peer_info_key in service_config:
        if peer_host is not None:
            service_config[peer_info_key]["host"] = peer_host
        if peer_port is not None:
            service_config[peer_info_key]["port"] = peer_port
    elif f"{peer_info_key}s" in service_config and len(service_config[f"{peer_info_key}s"]) > 0:
        if peer_host is not None:
            service_config[f"{peer_info_key}s"][0]["host"] = peer_host
        if peer_port is not None:
            service_config[f"{peer_info_key}s"][0]["port"] = peer_port



from __future__ import annotations

import json
import logging
from sqlite3 import Row
from typing import Optional, TypeVar, Union

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.db_wrapper import DBWrapper2, execute_fetchone
from chik.util.ints import uint32
from chik.wallet.lineage_proof import LineageProof
from chik.wallet.nft_wallet.nft_info import DEFAULT_STATUS, IN_TRANSACTION_STATUS, NFTCoinInfo

log = logging.getLogger(__name__)
_T_WalletNftStore = TypeVar("_T_WalletNftStore", bound="WalletNftStore")
REMOVE_BUFF_BLOCKS = 1000
NFT_COIN_INFO_COLUMNS = "nft_id, coin, lineage_proof, mint_height, status, full_puzzle, latest_height, minter_did"


def _to_nft_coin_info(row: Row) -> NFTCoinInfo:
    # nft_id, coin, lineage_proof, mint_height, status, full_puzzle, latest_height, minter_did
    return NFTCoinInfo(
        bytes32.from_hexstr(row[0]),
        Coin.from_json_dict(json.loads(row[1])),
        None if row[2] is None else LineageProof.from_json_dict(json.loads(row[2])),
        Program.from_bytes(row[5]),
        uint32(row[3]),
        None if row[7] is None else bytes32.from_hexstr(row[7]),
        uint32(row[6]) if row[6] is not None else uint32(0),
        row[4] == IN_TRANSACTION_STATUS,
    )


class WalletNftStore:
    """
    WalletNftStore keeps track of all user created NFTs and necessary smart-contract data
    """

    db_wrapper: DBWrapper2

    @classmethod
    async def create(cls: type[_T_WalletNftStore], db_wrapper: DBWrapper2) -> _T_WalletNftStore:
        self = cls()
        self.db_wrapper = db_wrapper
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            await conn.execute(
                "CREATE TABLE IF NOT EXISTS users_nfts("
                " nft_id text PRIMARY KEY,"
                " nft_coin_id text,"
                " wallet_id int,"
                " did_id text,"
                " coin text,"
                " lineage_proof text,"
                " mint_height bigint,"
                " status text,"
                " full_puzzle blob)"
            )
            await conn.execute("CREATE INDEX IF NOT EXISTS nft_coin_id on users_nfts(nft_coin_id)")
            await conn.execute("CREATE INDEX IF NOT EXISTS nft_wallet_id on users_nfts(wallet_id)")
            await conn.execute("CREATE INDEX IF NOT EXISTS nft_did_id on users_nfts(did_id)")
            try:
                # Add your new column on the top, otherwise it will not be created.
                await conn.execute("ALTER TABLE users_nfts ADD COLUMN minter_did text")
                # These are patched columns for resolving reorg issue
                await conn.execute("ALTER TABLE users_nfts ADD COLUMN removed_height bigint")
                await conn.execute("ALTER TABLE users_nfts ADD COLUMN latest_height bigint")
                await conn.execute("CREATE INDEX IF NOT EXISTS removed_nft_height on users_nfts(removed_height)")
                await conn.execute("CREATE INDEX IF NOT EXISTS latest_nft_height on users_nfts(latest_height)")
            except Exception:
                pass

        return self

    async def delete_nft_by_nft_id(self, nft_id: bytes32, height: uint32) -> bool:
        """Tries to mark a given NFT as deleted at specific height

        This is due to how re-org works
        Returns `True` if NFT was found and marked deleted or `False` if not."""
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            # Remove NFT in the users_nfts table
            cursor = await conn.execute(
                "UPDATE users_nfts SET removed_height=? WHERE nft_id=?", (int(height), nft_id.hex())
            )
            return cursor.rowcount > 0

    async def delete_nft_by_coin_id(self, coin_id: bytes32, height: uint32) -> bool:
        """Tries to mark a given NFT as deleted at specific height

        This is due to how re-org works
        Returns `True` if NFT was found and marked deleted or `False` if not."""
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            # Remove NFT in the users_nfts table
            cursor = await conn.execute(
                "UPDATE users_nfts SET removed_height=? WHERE nft_coin_id=?", (int(height), coin_id.hex())
            )
            if cursor.rowcount > 0:
                log.info("Deleted NFT with coin id: %s", coin_id.hex())
                return True
            log.warning("Couldn't find NFT with coin id to delete: %s", coin_id)
            return False

    async def update_pending_transaction(self, nft_coin_id: bytes32, pending_transaction: bool) -> bool:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            c = await conn.execute(
                "UPDATE users_nfts SET status=? WHERE nft_coin_id = ?",
                (IN_TRANSACTION_STATUS if pending_transaction else DEFAULT_STATUS, nft_coin_id.hex()),
            )
            return c.rowcount > 0

    async def save_nft(self, wallet_id: uint32, did_id: Optional[bytes32], nft_coin_info: NFTCoinInfo) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            columns = (
                "nft_id, nft_coin_id, wallet_id, did_id, coin, lineage_proof, mint_height, status, full_puzzle, "
                "minter_did, removed_height, latest_height"
            )
            await conn.execute(
                f"INSERT or REPLACE INTO users_nfts ({columns}) VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                (
                    nft_coin_info.nft_id.hex(),
                    nft_coin_info.coin.name().hex(),
                    int(wallet_id),
                    did_id.hex() if did_id else None,
                    json.dumps(nft_coin_info.coin.to_json_dict()),
                    (
                        json.dumps(nft_coin_info.lineage_proof.to_json_dict())
                        if nft_coin_info.lineage_proof is not None
                        else None
                    ),
                    int(nft_coin_info.mint_height),
                    IN_TRANSACTION_STATUS if nft_coin_info.pending_transaction else DEFAULT_STATUS,
                    bytes(nft_coin_info.full_puzzle),
                    None if nft_coin_info.minter_did is None else nft_coin_info.minter_did.hex(),
                    None,
                    int(nft_coin_info.latest_height),
                ),
            )
            # Rotate the old removed NFTs, they are not possible to be reorged
            await conn.execute(
                "DELETE FROM users_nfts WHERE removed_height is not NULL and removed_height<?",
                (int(nft_coin_info.latest_height) - REMOVE_BUFF_BLOCKS,),
            )

    async def count(self, wallet_id: Optional[uint32] = None, did_id: Optional[bytes32] = None) -> int:
        sql = "SELECT COUNT(nft_id) FROM users_nfts WHERE removed_height is NULL"
        params: list[Union[uint32, bytes32]] = []
        if wallet_id is not None:
            sql += " AND wallet_id=?"
            params.append(wallet_id)
        if did_id is not None:
            sql += " AND did_id=?"
            params.append(did_id)
        async with self.db_wrapper.reader_no_transaction() as conn:
            count_row = await execute_fetchone(conn, sql, params)
            if count_row:
                return int(count_row[0])
        return -1

    async def is_empty(self, wallet_id: Optional[uint32] = None) -> bool:
        sql = "SELECT 1 FROM users_nfts WHERE removed_height is NULL"
        params: list[Union[uint32, bytes32]] = []
        if wallet_id is not None:
            sql += " AND wallet_id=?"
            params.append(wallet_id)
        sql += " LIMIT 1"
        async with self.db_wrapper.reader_no_transaction() as conn:
            count_row = await execute_fetchone(conn, sql, params)
            if count_row:
                return False
        return True

    async def get_nft_list(
        self,
        wallet_id: Optional[uint32] = None,
        did_id: Optional[bytes32] = None,
        start_index: int = 0,
        count: int = 50,
    ) -> list[NFTCoinInfo]:
        try:
            start_index = int(start_index)
        except ValueError:
            start_index = 0
        try:
            count = int(count)
        except ValueError:
            count = 50

        sql: str = f"SELECT {NFT_COIN_INFO_COLUMNS} from users_nfts WHERE"
        if wallet_id is not None and did_id is None:
            sql += f" wallet_id={wallet_id}"
        if wallet_id is None and did_id is not None:
            sql += f" did_id='{did_id.hex()}'"
        if wallet_id is not None and did_id is not None:
            sql += f" did_id='{did_id.hex()}' and wallet_id={wallet_id}"
        if wallet_id is not None or did_id is not None:
            sql += " and"
        sql += " removed_height is NULL"
        sql += " LIMIT ? OFFSET ?"
        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await conn.execute_fetchall(sql, (count, start_index))

        return [
            NFTCoinInfo(
                bytes32.from_hexstr(row[0]),
                Coin.from_json_dict(json.loads(row[1])),
                None if row[2] is None else LineageProof.from_json_dict(json.loads(row[2])),
                Program.from_bytes(row[5]),
                uint32(row[3]),
                None if row[7] is None else bytes32.from_hexstr(row[7]),
                uint32(row[6]) if row[6] is not None else uint32(0),
                row[4] == IN_TRANSACTION_STATUS,
            )
            for row in rows
        ]

    async def exists(self, coin_id: bytes32) -> bool:
        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await execute_fetchone(
                conn,
                "SELECT EXISTS(SELECT nft_id"
                " from users_nfts WHERE removed_height is NULL and nft_coin_id=? LIMIT 1)",
                (coin_id.hex(),),
            )
            return True if rows and rows[0] == 1 else False

    async def get_nft_by_coin_id(self, nft_coin_id: bytes32) -> Optional[NFTCoinInfo]:
        async with self.db_wrapper.reader_no_transaction() as conn:
            rows = await conn.execute_fetchall(
                f"SELECT {NFT_COIN_INFO_COLUMNS} from users_nfts WHERE removed_height is NULL and nft_coin_id = ?",
                (nft_coin_id.hex(),),
            )
        rows = list(rows)
        if len(rows) == 1:
            return _to_nft_coin_info(rows[0])
        elif len(rows) == 2:
            raise ValueError("Can only return one NFT, but found > 1 from given coin ids")
        return None

    async def get_nft_by_id(self, nft_id: bytes32, wallet_id: Optional[uint32] = None) -> Optional[NFTCoinInfo]:
        async with self.db_wrapper.reader_no_transaction() as conn:
            sql = f"SELECT {NFT_COIN_INFO_COLUMNS} from users_nfts WHERE removed_height is NULL and nft_id=?"
            params: list[Union[uint32, str]] = [nft_id.hex()]
            if wallet_id:
                sql += " and wallet_id=?"
                params.append(wallet_id)
            row = await execute_fetchone(
                conn,
                sql,
                params,
            )

        if row is None:
            return None

        return _to_nft_coin_info(row)

    async def rollback_to_block(self, height: int) -> bool:
        """
        Rolls back the blockchain to block_index. All coins confirmed after this point are removed.
        All coins spent after this point are set to unspent. Can be -1 (rollback all)
        """
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            # Remove reorged NFTs
            await conn.execute("DELETE FROM users_nfts WHERE latest_height>?", (height,))

            # Retrieve removed NFTs
            result = await conn.execute(
                "UPDATE users_nfts SET removed_height = null WHERE removed_height>?",
                (height,),
            )
            if result.rowcount > 0:
                return True
            return False

    async def delete_wallet(self, wallet_id: uint32) -> None:
        async with self.db_wrapper.writer_maybe_transaction() as conn:
            cursor = await conn.execute("DELETE FROM users_nfts WHERE wallet_id=?", (wallet_id,))
            await cursor.close()



from __future__ import annotations

from collections.abc import AsyncIterator, Awaitable
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional

import pytest

from chik._tests.util.misc import Marks, RecordingWebServer, datacases
from chik.rpc.rpc_client import ResponseFailureError, RpcClient
from chik.rpc.rpc_server import RpcServer
from chik.server.outbound_message import NodeType
from chik.util.ints import uint16

non_fetch_client_methods = {
    RpcClient.create,
    RpcClient.create_as_context,
    RpcClient.fetch,
    RpcClient.close,
    RpcClient.await_closed,
}

client_fetch_methods = [
    attribute
    for name, attribute in vars(RpcClient).items()
    if callable(attribute) and attribute not in non_fetch_client_methods and not name.startswith("__")
]


@dataclass
class InvalidCreateCase:
    id: str
    root_path: Optional[Path] = None
    net_config: Optional[dict[str, Any]] = None
    marks: Marks = ()


@pytest.fixture(name="rpc_client")
async def rpc_client_fixture(recording_web_server: RecordingWebServer) -> AsyncIterator[RpcClient]:
    async with RpcClient.create_as_context(
        self_hostname=recording_web_server.web_server.hostname,
        port=recording_web_server.web_server.listen_port,
    ) as rpc_client:
        yield rpc_client


@datacases(
    InvalidCreateCase(id="just root path", root_path=Path("/root/path")),
    InvalidCreateCase(id="just net config", net_config={}),
)
@pytest.mark.anyio
async def test_rpc_client_create_raises_for_invalid_root_path_net_config_combinations(
    case: InvalidCreateCase,
) -> None:
    with pytest.raises(ValueError, match="Either both or neither of"):
        await RpcClient.create(
            self_hostname="",
            port=uint16(0),
            root_path=case.root_path,
            net_config=case.net_config,
        )


@pytest.mark.anyio
async def test_rpc_client_works_without_ssl(recording_web_server: RecordingWebServer) -> None:
    expected_result = {"success": True, "daddy": "putdown"}

    async with RpcClient.create_as_context(
        self_hostname=recording_web_server.web_server.hostname,
        port=recording_web_server.web_server.listen_port,
    ) as rpc_client:
        result = await rpc_client.fetch(path="", request_json={"response": expected_result})

    assert result == expected_result


@pytest.mark.anyio
async def test_rpc_client_send_request(
    rpc_client: RpcClient,
) -> None:
    expected_response = {"success": True, "magic": "asparagus"}

    response = await rpc_client.fetch(path="/table", request_json={"response": expected_response})

    assert response == expected_response


@pytest.mark.anyio
async def test_failure_exception(
    rpc_client: RpcClient,
) -> None:
    expected_response = {"success": False, "magic": "xylophone"}

    with pytest.raises(ResponseFailureError) as exception_info:
        await rpc_client.fetch(path="/table", request_json={"response": expected_response})

    assert exception_info.value.response == expected_response


def test_client_standard_endpoints_match_server() -> None:
    # NOTE: this test assumes that the client method names should match the server
    #       route names
    client_method_names = {method.__name__ for method in client_fetch_methods}
    server_route_names = {method.lstrip("/") for method in RpcServer._routes.keys()}
    assert client_method_names == server_route_names


@pytest.mark.anyio
@pytest.mark.parametrize("client_method", client_fetch_methods)
async def test_client_fetch_methods(
    client_method: Callable[..., Awaitable[object]],
    rpc_client: RpcClient,
    recording_web_server: RecordingWebServer,
) -> None:
    # NOTE: this test assumes that the client method names should match the server
    #       route names

    parameters: dict[Callable[..., Awaitable[object]], dict[str, object]] = {
        RpcClient.open_connection: {"host": "", "port": 0},
        RpcClient.close_connection: {"node_id": b""},
        RpcClient.get_connections: {"node_type": NodeType.FULL_NODE},
        RpcClient.set_log_level: {"level": "DEBUG"},
    }

    try:
        await client_method(rpc_client, **parameters.get(client_method, {}))
    except Exception as exception:
        if client_method is RpcClient.get_connections and isinstance(exception, KeyError):
            pass
        else:  # pragma: no cover
            # this case will fail the test so not normally executed
            raise

    [request] = recording_web_server.requests
    assert request.content_type == "application/json"
    assert request.method == "POST"
    assert request.path == f"/{client_method.__name__}"



from __future__ import annotations

from typing import Optional, Union

from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, compute_additions
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.util.curry_and_treehash import (
    calculate_hash_of_quoted_mod_hash,
    curry_and_treehash,
    shatree_atom,
    shatree_pair,
)

SINGLETON_TOP_LAYER_MOD = load_klvm_maybe_recompile("singleton_top_layer_v1_1.clsp")
SINGLETON_TOP_LAYER_MOD_HASH = SINGLETON_TOP_LAYER_MOD.get_tree_hash()
SINGLETON_TOP_LAYER_MOD_HASH_TREE_HASH = shatree_atom(SINGLETON_TOP_LAYER_MOD_HASH)
SINGLETON_TOP_LAYER_MOD_HASH_QUOTED = calculate_hash_of_quoted_mod_hash(SINGLETON_TOP_LAYER_MOD_HASH)
SINGLETON_LAUNCHER_PUZZLE = load_klvm_maybe_recompile("singleton_launcher.clsp")
SINGLETON_LAUNCHER_PUZZLE_HASH = SINGLETON_LAUNCHER_PUZZLE.get_tree_hash()
SINGLETON_LAUNCHER_PUZZLE_HASH_TREE_HASH = shatree_atom(SINGLETON_LAUNCHER_PUZZLE_HASH)


def get_inner_puzzle_from_singleton(puzzle: Union[Program, SerializedProgram]) -> Optional[Program]:
    """
    Extract the inner puzzle of a singleton
    :param puzzle: Singleton puzzle
    :return: Inner puzzle
    """
    r = puzzle.uncurry()
    if r is None:
        return None
    inner_f, args = r
    if not is_singleton(inner_f):
        return None
    _SINGLETON_STRUCT, INNER_PUZZLE = list(args.as_iter())
    return Program(INNER_PUZZLE)


def get_singleton_id_from_puzzle(puzzle: Union[Program, SerializedProgram]) -> Optional[bytes32]:
    """
    Extract the singleton ID from a singleton puzzle
    :param puzzle: Singleton puzzle
    :return: Inner puzzle
    """
    r = puzzle.uncurry()
    if r is None:
        return None  # pragma: no cover
    inner_f, args = r
    if not is_singleton(inner_f):
        return None
    SINGLETON_STRUCT, _INNER_PUZZLE = list(args.as_iter())
    return bytes32(Program(SINGLETON_STRUCT).rest().first().as_atom())


def is_singleton(inner_f: Union[Program, SerializedProgram]) -> bool:
    """
    Check if a puzzle is a singleton mod
    :param inner_f: puzzle
    :return: Boolean
    """
    return inner_f == SINGLETON_TOP_LAYER_MOD


def create_singleton_puzzle_hash(innerpuz_hash: bytes32, launcher_id: bytes32) -> bytes32:
    """
    Return Hash ID of the whole Singleton Puzzle
    :param innerpuz_hash: Singleton inner puzzle tree hash
    :param launcher_id: launcher coin name
    :return: Singleton full puzzle hash
    """
    # singleton_struct = (MOD_HASH . (LAUNCHER_ID . LAUNCHER_PUZZLE_HASH))
    singleton_struct = shatree_pair(
        SINGLETON_TOP_LAYER_MOD_HASH_TREE_HASH,
        shatree_pair(shatree_atom(launcher_id), SINGLETON_LAUNCHER_PUZZLE_HASH_TREE_HASH),
    )

    return curry_and_treehash(SINGLETON_TOP_LAYER_MOD_HASH_QUOTED, singleton_struct, innerpuz_hash)


def create_singleton_puzzle(innerpuz: Union[Program, SerializedProgram], launcher_id: bytes32) -> Program:
    """
    Create a full Singleton puzzle
    :param innerpuz: Singleton inner puzzle
    :param launcher_id:
    :return: Singleton full puzzle
    """
    # singleton_struct = (MOD_HASH . (LAUNCHER_ID . LAUNCHER_PUZZLE_HASH))
    singleton_struct = Program.to((SINGLETON_TOP_LAYER_MOD_HASH, (launcher_id, SINGLETON_LAUNCHER_PUZZLE_HASH)))
    return SINGLETON_TOP_LAYER_MOD.curry(singleton_struct, innerpuz)


def get_most_recent_singleton_coin_from_coin_spend(coin_sol: CoinSpend) -> Optional[Coin]:
    additions: list[Coin] = compute_additions(coin_sol)
    for coin in additions:
        if coin.amount % 2 == 1:
            return coin
    return None  # pragma: no cover


def get_singleton_struct_for_id(id: bytes32) -> Program:
    singleton_struct: Program = Program.to((SINGLETON_TOP_LAYER_MOD_HASH, (id, SINGLETON_LAUNCHER_PUZZLE_HASH)))
    return singleton_struct



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Optional

from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.ints import uint16
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.puzzles.singleton_top_layer_v1_1 import SINGLETON_LAUNCHER_HASH, SINGLETON_MOD_HASH
from chik.wallet.uncurried_puzzle import UncurriedPuzzle

TRANSFER_PROGRAM_MOD = load_klvm_maybe_recompile(
    "nft_ownership_transfer_program_one_way_claim_with_royalties.clsp",
    package_or_requirement="chik.wallet.nft_wallet.puzzles",
)


def match_transfer_program_puzzle(puzzle: UncurriedPuzzle) -> tuple[bool, list[Program]]:
    if puzzle.mod == TRANSFER_PROGRAM_MOD:
        return True, list(puzzle.args.as_iter())
    return False, []


def puzzle_for_transfer_program(launcher_id: bytes32, royalty_puzzle_hash: bytes32, percentage: uint16) -> Program:
    singleton_struct = Program.to((SINGLETON_MOD_HASH, (launcher_id, SINGLETON_LAUNCHER_HASH)))
    return TRANSFER_PROGRAM_MOD.curry(
        singleton_struct,
        royalty_puzzle_hash,
        percentage,
    )


def solution_for_transfer_program(
    conditions: Program,
    current_owner: Optional[bytes32],
    new_did: bytes32,
    new_did_inner_hash: bytes32,
    trade_prices_list: Program,
) -> Program:
    return Program.to([conditions, current_owner, [new_did, trade_prices_list, new_did_inner_hash]])


@dataclass(frozen=True)
class TransferProgramPuzzle:
    _match: Callable[[UncurriedPuzzle], Optional[PuzzleInfo]]
    _construct: Callable[[PuzzleInfo, Program], Program]
    _solve: Callable[[PuzzleInfo, Solver, Program, Program], Program]
    _get_inner_puzzle: Callable[[PuzzleInfo, UncurriedPuzzle], Optional[Program]]
    _get_inner_solution: Callable[[PuzzleInfo, Program], Optional[Program]]

    def match(self, puzzle: UncurriedPuzzle) -> Optional[PuzzleInfo]:
        matched, curried_args = match_transfer_program_puzzle(puzzle)
        if matched:
            singleton_struct, royalty_puzzle_hash, percentage = curried_args
            constructor_dict = {
                "type": "royalty transfer program",
                "launcher_id": "0x" + singleton_struct.rest().first().as_python().hex(),
                "royalty_address": "0x" + royalty_puzzle_hash.as_python().hex(),
                "royalty_percentage": str(percentage.as_int()),
            }
            return PuzzleInfo(constructor_dict)
        else:
            return None

    def asset_id(self, constructor: PuzzleInfo) -> Optional[bytes32]:
        return None

    def construct(self, constructor: PuzzleInfo, inner_puzzle: Program) -> Program:
        return puzzle_for_transfer_program(
            constructor["launcher_id"], constructor["royalty_address"], constructor["royalty_percentage"]
        )

    def get_inner_puzzle(self, constructor: PuzzleInfo, puzzle_reveal: UncurriedPuzzle) -> Optional[Program]:
        return None

    def get_inner_solution(self, constructor: PuzzleInfo, solution: Program) -> Optional[Program]:
        return None

    def solve(self, constructor: PuzzleInfo, solver: Solver, inner_puzzle: Program, inner_solution: Program) -> Program:
        return Program.to(None)



# ruff: noqa: F405
# this file is generated by build_network_protocol_files.py

from __future__ import annotations

from chik._tests.util.network_protocol_data import *  # noqa: F403
from chik._tests.util.protocol_messages_json import *  # noqa: F403


def test_protocol_json() -> None:
    assert str(new_signage_point_json) == str(new_signage_point.to_json_dict())
    assert type(new_signage_point).from_json_dict(new_signage_point_json) == new_signage_point
    assert str(declare_proof_of_space_json) == str(declare_proof_of_space.to_json_dict())
    assert type(declare_proof_of_space).from_json_dict(declare_proof_of_space_json) == declare_proof_of_space
    assert str(request_signed_values_json) == str(request_signed_values.to_json_dict())
    assert type(request_signed_values).from_json_dict(request_signed_values_json) == request_signed_values
    assert str(farming_info_json) == str(farming_info.to_json_dict())
    assert type(farming_info).from_json_dict(farming_info_json) == farming_info
    assert str(signed_values_json) == str(signed_values.to_json_dict())
    assert type(signed_values).from_json_dict(signed_values_json) == signed_values
    assert str(new_peak_json) == str(new_peak.to_json_dict())
    assert type(new_peak).from_json_dict(new_peak_json) == new_peak
    assert str(new_transaction_json) == str(new_transaction.to_json_dict())
    assert type(new_transaction).from_json_dict(new_transaction_json) == new_transaction
    assert str(request_transaction_json) == str(request_transaction.to_json_dict())
    assert type(request_transaction).from_json_dict(request_transaction_json) == request_transaction
    assert str(respond_transaction_json) == str(respond_transaction.to_json_dict())
    assert type(respond_transaction).from_json_dict(respond_transaction_json) == respond_transaction
    assert str(request_proof_of_weight_json) == str(request_proof_of_weight.to_json_dict())
    assert type(request_proof_of_weight).from_json_dict(request_proof_of_weight_json) == request_proof_of_weight
    assert str(respond_proof_of_weight_json) == str(respond_proof_of_weight.to_json_dict())
    assert type(respond_proof_of_weight).from_json_dict(respond_proof_of_weight_json) == respond_proof_of_weight
    assert str(request_block_json) == str(request_block.to_json_dict())
    assert type(request_block).from_json_dict(request_block_json) == request_block
    assert str(reject_block_json) == str(reject_block.to_json_dict())
    assert type(reject_block).from_json_dict(reject_block_json) == reject_block
    assert str(request_blocks_json) == str(request_blocks.to_json_dict())
    assert type(request_blocks).from_json_dict(request_blocks_json) == request_blocks
    assert str(respond_blocks_json) == str(respond_blocks.to_json_dict())
    assert type(respond_blocks).from_json_dict(respond_blocks_json) == respond_blocks
    assert str(reject_blocks_json) == str(reject_blocks.to_json_dict())
    assert type(reject_blocks).from_json_dict(reject_blocks_json) == reject_blocks
    assert str(respond_block_json) == str(respond_block.to_json_dict())
    assert type(respond_block).from_json_dict(respond_block_json) == respond_block
    assert str(new_unfinished_block_json) == str(new_unfinished_block.to_json_dict())
    assert type(new_unfinished_block).from_json_dict(new_unfinished_block_json) == new_unfinished_block
    assert str(request_unfinished_block_json) == str(request_unfinished_block.to_json_dict())
    assert type(request_unfinished_block).from_json_dict(request_unfinished_block_json) == request_unfinished_block
    assert str(respond_unfinished_block_json) == str(respond_unfinished_block.to_json_dict())
    assert type(respond_unfinished_block).from_json_dict(respond_unfinished_block_json) == respond_unfinished_block
    assert str(new_signage_point_or_end_of_subslot_json) == str(new_signage_point_or_end_of_subslot.to_json_dict())
    assert (
        type(new_signage_point_or_end_of_subslot).from_json_dict(new_signage_point_or_end_of_subslot_json)
        == new_signage_point_or_end_of_subslot
    )
    assert str(request_signage_point_or_end_of_subslot_json) == str(
        request_signage_point_or_end_of_subslot.to_json_dict()
    )
    assert (
        type(request_signage_point_or_end_of_subslot).from_json_dict(request_signage_point_or_end_of_subslot_json)
        == request_signage_point_or_end_of_subslot
    )
    assert str(respond_signage_point_json) == str(respond_signage_point.to_json_dict())
    assert type(respond_signage_point).from_json_dict(respond_signage_point_json) == respond_signage_point
    assert str(respond_end_of_subslot_json) == str(respond_end_of_subslot.to_json_dict())
    assert type(respond_end_of_subslot).from_json_dict(respond_end_of_subslot_json) == respond_end_of_subslot
    assert str(request_mempool_transaction_json) == str(request_mempool_transaction.to_json_dict())
    assert (
        type(request_mempool_transaction).from_json_dict(request_mempool_transaction_json)
        == request_mempool_transaction
    )
    assert str(new_compact_vdf_json) == str(new_compact_vdf.to_json_dict())
    assert type(new_compact_vdf).from_json_dict(new_compact_vdf_json) == new_compact_vdf
    assert str(request_compact_vdf_json) == str(request_compact_vdf.to_json_dict())
    assert type(request_compact_vdf).from_json_dict(request_compact_vdf_json) == request_compact_vdf
    assert str(respond_compact_vdf_json) == str(respond_compact_vdf.to_json_dict())
    assert type(respond_compact_vdf).from_json_dict(respond_compact_vdf_json) == respond_compact_vdf
    assert str(request_peers_json) == str(request_peers.to_json_dict())
    assert type(request_peers).from_json_dict(request_peers_json) == request_peers
    assert str(respond_peers_json) == str(respond_peers.to_json_dict())
    assert type(respond_peers).from_json_dict(respond_peers_json) == respond_peers
    assert str(new_unfinished_block2_json) == str(new_unfinished_block2.to_json_dict())
    assert type(new_unfinished_block2).from_json_dict(new_unfinished_block2_json) == new_unfinished_block2
    assert str(request_unfinished_block2_json) == str(request_unfinished_block2.to_json_dict())
    assert type(request_unfinished_block2).from_json_dict(request_unfinished_block2_json) == request_unfinished_block2
    assert str(request_puzzle_solution_json) == str(request_puzzle_solution.to_json_dict())
    assert type(request_puzzle_solution).from_json_dict(request_puzzle_solution_json) == request_puzzle_solution
    assert str(puzzle_solution_response_json) == str(puzzle_solution_response.to_json_dict())
    assert type(puzzle_solution_response).from_json_dict(puzzle_solution_response_json) == puzzle_solution_response
    assert str(respond_puzzle_solution_json) == str(respond_puzzle_solution.to_json_dict())
    assert type(respond_puzzle_solution).from_json_dict(respond_puzzle_solution_json) == respond_puzzle_solution
    assert str(reject_puzzle_solution_json) == str(reject_puzzle_solution.to_json_dict())
    assert type(reject_puzzle_solution).from_json_dict(reject_puzzle_solution_json) == reject_puzzle_solution
    assert str(send_transaction_json) == str(send_transaction.to_json_dict())
    assert type(send_transaction).from_json_dict(send_transaction_json) == send_transaction
    assert str(transaction_ack_json) == str(transaction_ack.to_json_dict())
    assert type(transaction_ack).from_json_dict(transaction_ack_json) == transaction_ack
    assert str(new_peak_wallet_json) == str(new_peak_wallet.to_json_dict())
    assert type(new_peak_wallet).from_json_dict(new_peak_wallet_json) == new_peak_wallet
    assert str(request_block_header_json) == str(request_block_header.to_json_dict())
    assert type(request_block_header).from_json_dict(request_block_header_json) == request_block_header
    assert str(request_block_headers_json) == str(request_block_headers.to_json_dict())
    assert type(request_block_headers).from_json_dict(request_block_headers_json) == request_block_headers
    assert str(respond_header_block_json) == str(respond_header_block.to_json_dict())
    assert type(respond_header_block).from_json_dict(respond_header_block_json) == respond_header_block
    assert str(respond_block_headers_json) == str(respond_block_headers.to_json_dict())
    assert type(respond_block_headers).from_json_dict(respond_block_headers_json) == respond_block_headers
    assert str(reject_header_request_json) == str(reject_header_request.to_json_dict())
    assert type(reject_header_request).from_json_dict(reject_header_request_json) == reject_header_request
    assert str(request_removals_json) == str(request_removals.to_json_dict())
    assert type(request_removals).from_json_dict(request_removals_json) == request_removals
    assert str(respond_removals_json) == str(respond_removals.to_json_dict())
    assert type(respond_removals).from_json_dict(respond_removals_json) == respond_removals
    assert str(reject_removals_request_json) == str(reject_removals_request.to_json_dict())
    assert type(reject_removals_request).from_json_dict(reject_removals_request_json) == reject_removals_request
    assert str(request_additions_json) == str(request_additions.to_json_dict())
    assert type(request_additions).from_json_dict(request_additions_json) == request_additions
    assert str(respond_additions_json) == str(respond_additions.to_json_dict())
    assert type(respond_additions).from_json_dict(respond_additions_json) == respond_additions
    assert str(reject_additions_json) == str(reject_additions.to_json_dict())
    assert type(reject_additions).from_json_dict(reject_additions_json) == reject_additions
    assert str(request_header_blocks_json) == str(request_header_blocks.to_json_dict())
    assert type(request_header_blocks).from_json_dict(request_header_blocks_json) == request_header_blocks
    assert str(reject_header_blocks_json) == str(reject_header_blocks.to_json_dict())
    assert type(reject_header_blocks).from_json_dict(reject_header_blocks_json) == reject_header_blocks
    assert str(respond_header_blocks_json) == str(respond_header_blocks.to_json_dict())
    assert type(respond_header_blocks).from_json_dict(respond_header_blocks_json) == respond_header_blocks
    assert str(coin_state_json) == str(coin_state.to_json_dict())
    assert type(coin_state).from_json_dict(coin_state_json) == coin_state
    assert str(register_for_ph_updates_json) == str(register_for_ph_updates.to_json_dict())
    assert type(register_for_ph_updates).from_json_dict(register_for_ph_updates_json) == register_for_ph_updates
    assert str(reject_block_headers_json) == str(reject_block_headers.to_json_dict())
    assert type(reject_block_headers).from_json_dict(reject_block_headers_json) == reject_block_headers
    assert str(respond_to_ph_updates_json) == str(respond_to_ph_updates.to_json_dict())
    assert type(respond_to_ph_updates).from_json_dict(respond_to_ph_updates_json) == respond_to_ph_updates
    assert str(register_for_coin_updates_json) == str(register_for_coin_updates.to_json_dict())
    assert type(register_for_coin_updates).from_json_dict(register_for_coin_updates_json) == register_for_coin_updates
    assert str(respond_to_coin_updates_json) == str(respond_to_coin_updates.to_json_dict())
    assert type(respond_to_coin_updates).from_json_dict(respond_to_coin_updates_json) == respond_to_coin_updates
    assert str(coin_state_update_json) == str(coin_state_update.to_json_dict())
    assert type(coin_state_update).from_json_dict(coin_state_update_json) == coin_state_update
    assert str(request_children_json) == str(request_children.to_json_dict())
    assert type(request_children).from_json_dict(request_children_json) == request_children
    assert str(respond_children_json) == str(respond_children.to_json_dict())
    assert type(respond_children).from_json_dict(respond_children_json) == respond_children
    assert str(request_ses_info_json) == str(request_ses_info.to_json_dict())
    assert type(request_ses_info).from_json_dict(request_ses_info_json) == request_ses_info
    assert str(respond_ses_info_json) == str(respond_ses_info.to_json_dict())
    assert type(respond_ses_info).from_json_dict(respond_ses_info_json) == respond_ses_info
    assert str(coin_state_filters_json) == str(coin_state_filters.to_json_dict())
    assert type(coin_state_filters).from_json_dict(coin_state_filters_json) == coin_state_filters
    assert str(request_remove_puzzle_subscriptions_json) == str(request_remove_puzzle_subscriptions.to_json_dict())
    assert (
        type(request_remove_puzzle_subscriptions).from_json_dict(request_remove_puzzle_subscriptions_json)
        == request_remove_puzzle_subscriptions
    )
    assert str(respond_remove_puzzle_subscriptions_json) == str(respond_remove_puzzle_subscriptions.to_json_dict())
    assert (
        type(respond_remove_puzzle_subscriptions).from_json_dict(respond_remove_puzzle_subscriptions_json)
        == respond_remove_puzzle_subscriptions
    )
    assert str(request_remove_coin_subscriptions_json) == str(request_remove_coin_subscriptions.to_json_dict())
    assert (
        type(request_remove_coin_subscriptions).from_json_dict(request_remove_coin_subscriptions_json)
        == request_remove_coin_subscriptions
    )
    assert str(respond_remove_coin_subscriptions_json) == str(respond_remove_coin_subscriptions.to_json_dict())
    assert (
        type(respond_remove_coin_subscriptions).from_json_dict(respond_remove_coin_subscriptions_json)
        == respond_remove_coin_subscriptions
    )
    assert str(request_puzzle_state_json) == str(request_puzzle_state.to_json_dict())
    assert type(request_puzzle_state).from_json_dict(request_puzzle_state_json) == request_puzzle_state
    assert str(reject_puzzle_state_json) == str(reject_puzzle_state.to_json_dict())
    assert type(reject_puzzle_state).from_json_dict(reject_puzzle_state_json) == reject_puzzle_state
    assert str(respond_puzzle_state_json) == str(respond_puzzle_state.to_json_dict())
    assert type(respond_puzzle_state).from_json_dict(respond_puzzle_state_json) == respond_puzzle_state
    assert str(request_coin_state_json) == str(request_coin_state.to_json_dict())
    assert type(request_coin_state).from_json_dict(request_coin_state_json) == request_coin_state
    assert str(respond_coin_state_json) == str(respond_coin_state.to_json_dict())
    assert type(respond_coin_state).from_json_dict(respond_coin_state_json) == respond_coin_state
    assert str(reject_coin_state_json) == str(reject_coin_state.to_json_dict())
    assert type(reject_coin_state).from_json_dict(reject_coin_state_json) == reject_coin_state
    assert str(request_cost_info_json) == str(request_cost_info.to_json_dict())
    assert type(request_cost_info).from_json_dict(request_cost_info_json) == request_cost_info
    assert str(respond_cost_info_json) == str(respond_cost_info.to_json_dict())
    assert type(respond_cost_info).from_json_dict(respond_cost_info_json) == respond_cost_info
    assert str(pool_difficulty_json) == str(pool_difficulty.to_json_dict())
    assert type(pool_difficulty).from_json_dict(pool_difficulty_json) == pool_difficulty
    assert str(harvester_handhsake_json) == str(harvester_handhsake.to_json_dict())
    assert type(harvester_handhsake).from_json_dict(harvester_handhsake_json) == harvester_handhsake
    assert str(new_signage_point_harvester_json) == str(new_signage_point_harvester.to_json_dict())
    assert (
        type(new_signage_point_harvester).from_json_dict(new_signage_point_harvester_json)
        == new_signage_point_harvester
    )
    assert str(new_proof_of_space_json) == str(new_proof_of_space.to_json_dict())
    assert type(new_proof_of_space).from_json_dict(new_proof_of_space_json) == new_proof_of_space
    assert str(request_signatures_json) == str(request_signatures.to_json_dict())
    assert type(request_signatures).from_json_dict(request_signatures_json) == request_signatures
    assert str(respond_signatures_json) == str(respond_signatures.to_json_dict())
    assert type(respond_signatures).from_json_dict(respond_signatures_json) == respond_signatures
    assert str(plot_json) == str(plot.to_json_dict())
    assert type(plot).from_json_dict(plot_json) == plot
    assert str(request_plots_json) == str(request_plots.to_json_dict())
    assert type(request_plots).from_json_dict(request_plots_json) == request_plots
    assert str(respond_plots_json) == str(respond_plots.to_json_dict())
    assert type(respond_plots).from_json_dict(respond_plots_json) == respond_plots
    assert str(request_peers_introducer_json) == str(request_peers_introducer.to_json_dict())
    assert type(request_peers_introducer).from_json_dict(request_peers_introducer_json) == request_peers_introducer
    assert str(respond_peers_introducer_json) == str(respond_peers_introducer.to_json_dict())
    assert type(respond_peers_introducer).from_json_dict(respond_peers_introducer_json) == respond_peers_introducer
    assert str(authentication_payload_json) == str(authentication_payload.to_json_dict())
    assert type(authentication_payload).from_json_dict(authentication_payload_json) == authentication_payload
    assert str(get_pool_info_response_json) == str(get_pool_info_response.to_json_dict())
    assert type(get_pool_info_response).from_json_dict(get_pool_info_response_json) == get_pool_info_response
    assert str(post_partial_payload_json) == str(post_partial_payload.to_json_dict())
    assert type(post_partial_payload).from_json_dict(post_partial_payload_json) == post_partial_payload
    assert str(post_partial_request_json) == str(post_partial_request.to_json_dict())
    assert type(post_partial_request).from_json_dict(post_partial_request_json) == post_partial_request
    assert str(post_partial_response_json) == str(post_partial_response.to_json_dict())
    assert type(post_partial_response).from_json_dict(post_partial_response_json) == post_partial_response
    assert str(get_farmer_response_json) == str(get_farmer_response.to_json_dict())
    assert type(get_farmer_response).from_json_dict(get_farmer_response_json) == get_farmer_response
    assert str(post_farmer_payload_json) == str(post_farmer_payload.to_json_dict())
    assert type(post_farmer_payload).from_json_dict(post_farmer_payload_json) == post_farmer_payload
    assert str(post_farmer_request_json) == str(post_farmer_request.to_json_dict())
    assert type(post_farmer_request).from_json_dict(post_farmer_request_json) == post_farmer_request
    assert str(post_farmer_response_json) == str(post_farmer_response.to_json_dict())
    assert type(post_farmer_response).from_json_dict(post_farmer_response_json) == post_farmer_response
    assert str(put_farmer_payload_json) == str(put_farmer_payload.to_json_dict())
    assert type(put_farmer_payload).from_json_dict(put_farmer_payload_json) == put_farmer_payload
    assert str(put_farmer_request_json) == str(put_farmer_request.to_json_dict())
    assert type(put_farmer_request).from_json_dict(put_farmer_request_json) == put_farmer_request
    assert str(put_farmer_response_json) == str(put_farmer_response.to_json_dict())
    assert type(put_farmer_response).from_json_dict(put_farmer_response_json) == put_farmer_response
    assert str(error_response_json) == str(error_response.to_json_dict())
    assert type(error_response).from_json_dict(error_response_json) == error_response
    assert str(new_peak_timelord_json) == str(new_peak_timelord.to_json_dict())
    assert type(new_peak_timelord).from_json_dict(new_peak_timelord_json) == new_peak_timelord
    assert str(new_unfinished_block_timelord_json) == str(new_unfinished_block_timelord.to_json_dict())
    assert (
        type(new_unfinished_block_timelord).from_json_dict(new_unfinished_block_timelord_json)
        == new_unfinished_block_timelord
    )
    assert str(new_infusion_point_vdf_json) == str(new_infusion_point_vdf.to_json_dict())
    assert type(new_infusion_point_vdf).from_json_dict(new_infusion_point_vdf_json) == new_infusion_point_vdf
    assert str(new_signage_point_vdf_json) == str(new_signage_point_vdf.to_json_dict())
    assert type(new_signage_point_vdf).from_json_dict(new_signage_point_vdf_json) == new_signage_point_vdf
    assert str(new_end_of_sub_slot_bundle_json) == str(new_end_of_sub_slot_bundle.to_json_dict())
    assert (
        type(new_end_of_sub_slot_bundle).from_json_dict(new_end_of_sub_slot_bundle_json) == new_end_of_sub_slot_bundle
    )
    assert str(request_compact_proof_of_time_json) == str(request_compact_proof_of_time.to_json_dict())
    assert (
        type(request_compact_proof_of_time).from_json_dict(request_compact_proof_of_time_json)
        == request_compact_proof_of_time
    )
    assert str(respond_compact_proof_of_time_json) == str(respond_compact_proof_of_time.to_json_dict())
    assert (
        type(respond_compact_proof_of_time).from_json_dict(respond_compact_proof_of_time_json)
        == respond_compact_proof_of_time
    )
    assert str(error_without_data_json) == str(error_without_data.to_json_dict())
    assert type(error_without_data).from_json_dict(error_without_data_json) == error_without_data
    assert str(error_with_data_json) == str(error_with_data.to_json_dict())
    assert type(error_with_data).from_json_dict(error_with_data_json) == error_with_data



from __future__ import annotations

import asyncio
import logging
import os
import sys
import traceback
from pathlib import Path
from typing import Any, Optional

from chik.plotters.plotters_util import get_venv_bin, reset_loop_policy_for_windows, run_command, run_plotter
from chik.plotting.create_plots import resolve_plot_keys

log = logging.getLogger(__name__)


MADMAX_PLOTTER_DIR = "madmax-plotter"


def is_madmax_supported() -> bool:
    return sys.platform.startswith("linux") or sys.platform in {"darwin", "win32", "cygwin"}


def get_madmax_src_path(plotters_root_path: Path) -> Path:
    return plotters_root_path / MADMAX_PLOTTER_DIR


def get_madmax_package_path() -> Path:
    p = Path(os.path.dirname(sys.executable)).joinpath("_internal/madmax")
    if p.exists():
        return p
    return Path(os.path.dirname(sys.executable)).joinpath("madmax")


def get_madmax_exec_venv_path(ksize: int = 32) -> Optional[Path]:
    venv_bin_path = get_venv_bin()
    if not venv_bin_path:
        return None
    madmax_exec = "chik_plot"
    if ksize > 32:
        madmax_exec += "_k34"  # Use the chik_plot_k34 executable for k-sizes > 32
    if sys.platform in {"win32", "cygwin"}:
        madmax_exec += ".exe"
    return venv_bin_path / madmax_exec


def get_madmax_exec_src_path(plotters_root_path: Path, ksize: int = 32) -> Path:
    madmax_src_dir = get_madmax_src_path(plotters_root_path) / "build"
    madmax_exec = "chik_plot"
    if ksize > 32:
        madmax_exec += "_k34"  # Use the chik_plot_k34 executable for k-sizes > 32
    if sys.platform in {"win32", "cygwin"}:
        madmax_exec += ".exe"
    return madmax_src_dir / madmax_exec


def get_madmax_exec_package_path(ksize: int = 32) -> Path:
    madmax_dir: Path = get_madmax_package_path()
    madmax_exec: str = "chik_plot"
    if ksize > 32:
        madmax_exec += "_k34"  # Use the chik_plot_k34 executable for k-sizes > 32
    if sys.platform in {"win32", "cygwin"}:
        madmax_exec += ".exe"
    return madmax_dir / madmax_exec


def get_madmax_executable_path_for_ksize(plotters_root_path: Path, ksize: int = 32) -> Path:
    madmax_exec_venv_path = get_madmax_exec_venv_path(ksize)
    if madmax_exec_venv_path is not None and madmax_exec_venv_path.exists():
        return madmax_exec_venv_path
    madmax_exec_src_path = get_madmax_exec_src_path(plotters_root_path, ksize)
    if madmax_exec_src_path.exists():
        return madmax_exec_src_path
    return get_madmax_exec_package_path(ksize)


def get_madmax_version(plotters_root_path: Path):
    madmax_executable_path = get_madmax_executable_path_for_ksize(plotters_root_path)
    if not madmax_executable_path.exists():
        # (NotFound, "")
        return False, ""

    try:
        proc = run_command(
            [os.fspath(madmax_executable_path), "--version"],
            "Failed to call madmax with --version option",
            capture_output=True,
            text=True,
            check=False,
        )
        if proc.returncode != 0:
            return None, proc.stderr.strip()

        # (Found, versionStr)
        version_str = proc.stdout.strip()
        return True, version_str.split(".")
    except Exception as e:
        tb = traceback.format_exc()
        # (Unknown, ErrMsg)
        return None, f"Failed to determine madmax version: {e} {tb}"


def get_madmax_install_info(plotters_root_path: Path) -> Optional[dict[str, Any]]:
    info: dict[str, Any] = {"display_name": "madMAx Plotter"}
    installed: bool = False
    supported: bool = is_madmax_supported()

    madmax_executable_path = get_madmax_executable_path_for_ksize(plotters_root_path)
    if madmax_executable_path.exists():
        version: Optional[str] = None
        found, result_msg = get_madmax_version(plotters_root_path)
        if found:
            version = ".".join(result_msg)
        elif found is None:
            print(f"Failed to determine madMAx version: {result_msg}")

        if version is not None:
            installed = True
            info["version"] = version
        else:
            installed = False

    info["installed"] = installed
    if installed is False:
        info["can_install"] = supported

    return info


progress = {
    "[P1] Table 1 took": 0.01,
    "[P1] Table 2 took": 0.06,
    "[P1] Table 3 took": 0.12,
    "[P1] Table 4 took": 0.2,
    "[P1] Table 5 took": 0.28,
    "[P1] Table 6 took": 0.36,
    "[P1] Table 7 took": 0.42,
    "[P2] Table 7 rewrite took": 0.43,
    "[P2] Table 6 rewrite took": 0.48,
    "[P2] Table 5 rewrite took": 0.51,
    "[P2] Table 4 rewrite took": 0.55,
    "[P2] Table 3 rewrite took": 0.58,
    "[P2] Table 2 rewrite took": 0.61,
    "[P3-2] Table 2 took": 0.66,
    "[P3-2] Table 3 took": 0.73,
    "[P3-2] Table 4 took": 0.79,
    "[P3-2] Table 5 took": 0.85,
    "[P3-2] Table 6 took": 0.92,
    "[P3-2] Table 7 took": 0.98,
}


def dir_with_trailing_slash(dir: str) -> str:
    return dir if dir[-1] == os.path.sep else dir + os.path.sep


def plot_madmax(args, chik_root_path: Path, plotters_root_path: Path):
    if sys.platform != "win32" and sys.platform != "cygwin":
        import resource

        # madMAx has a ulimit -n requirement > 296:
        # "Cannot open at least 296 files, please raise maximum open file limit in OS."
        _soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
        # Set soft limit to max (hard limit)
        resource.setrlimit(resource.RLIMIT_NOFILE, (hard_limit, hard_limit))
    else:
        reset_loop_policy_for_windows()

    madmax_executable_path_for_ksize = get_madmax_executable_path_for_ksize(plotters_root_path, args.size)
    if not os.path.exists(madmax_executable_path_for_ksize):
        print("madmax plotter was not found.")
        return

    plot_keys = asyncio.run(
        resolve_plot_keys(
            None if args.farmerkey == b"" else args.farmerkey.hex(),
            None,
            None if args.pool_key == b"" else args.pool_key.hex(),
            None if args.contract == "" else args.contract,
            chik_root_path,
            log,
            args.connect_to_daemon,
        )
    )
    call_args = []
    call_args.append(os.fspath(madmax_executable_path_for_ksize))
    call_args.append("-f")
    call_args.append(bytes(plot_keys.farmer_public_key).hex())
    if plot_keys.pool_public_key is not None:
        call_args.append("-p")
        call_args.append(bytes(plot_keys.pool_public_key).hex())
    call_args.append("-t")
    # s if s[-1] == os.path.sep else s + os.path.sep
    call_args.append(dir_with_trailing_slash(args.tmpdir))
    if len(args.tmpdir2) > 0:
        call_args.append("-2")
        call_args.append(dir_with_trailing_slash(args.tmpdir2))
    call_args.append("-d")
    call_args.append(dir_with_trailing_slash(args.finaldir))
    if plot_keys.pool_contract_address is not None:
        call_args.append("-c")
        call_args.append(plot_keys.pool_contract_address)
    call_args.append("-n")
    call_args.append(str(args.count))
    call_args.append("-r")
    call_args.append(str(args.threads))
    call_args.append("-u")
    call_args.append(str(args.buckets))
    call_args.append("-v")
    call_args.append(str(args.buckets3))
    if args.waitforcopy:
        call_args.append("-w")
    if args.tmptoggle:
        call_args.append("-G")
    call_args.append("-K")
    call_args.append(str(args.rmulti2))
    if args.size != 32:
        call_args.append("-k")
        call_args.append(str(args.size))
    try:
        asyncio.run(run_plotter(chik_root_path, args.plotter, call_args, progress))
    except Exception as e:
        print(f"Exception while plotting: {type(e)} {e}")
        print(f"Traceback: {traceback.format_exc()}")



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, BinaryIO, Optional, Union

from chik_rs import G2Element
from klvm_tools.binutils import disassemble

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.coin import Coin, coin_as_list
from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend, make_spend
from chik.util.bech32m import bech32_decode, bech32_encode, convertbits
from chik.util.errors import Err, ValidationError
from chik.util.ints import uint64
from chik.util.streamable import parse_rust
from chik.wallet.conditions import (
    AssertCoinAnnouncement,
    AssertPuzzleAnnouncement,
    Condition,
    ConditionValidTimes,
    parse_conditions_non_consensus,
    parse_timelock_info,
)
from chik.wallet.outer_puzzles import (
    construct_puzzle,
    create_asset_id,
    get_inner_puzzle,
    get_inner_solution,
    match_puzzle,
    solve_puzzle,
)
from chik.wallet.payment import Payment
from chik.wallet.puzzle_drivers import PuzzleInfo, Solver
from chik.wallet.puzzles.load_klvm import load_klvm_maybe_recompile
from chik.wallet.uncurried_puzzle import UncurriedPuzzle, uncurry_puzzle
from chik.wallet.util.compute_hints import compute_spend_hints_and_additions
from chik.wallet.util.puzzle_compression import (
    compress_object_with_puzzles,
    decompress_object_with_puzzles,
    lowest_best_version,
)
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

OFFER_MOD = load_klvm_maybe_recompile("settlement_payments.clsp")
OFFER_MOD_HASH = OFFER_MOD.get_tree_hash()


def detect_dependent_coin(
    names: list[bytes32], deps: dict[bytes32, list[bytes32]], announcement_dict: dict[bytes32, list[bytes32]]
) -> Optional[tuple[bytes32, bytes32]]:
    # First, we check for any dependencies on coins in the same bundle
    for name in names:
        for dependency in deps[name]:
            for coin, announces in announcement_dict.items():
                if dependency in announces and coin != name:
                    # We found one, now remove it and anything that depends on it (except the "provider")
                    return name, coin
    return None


@dataclass(frozen=True)
class NotarizedPayment(Payment):
    nonce: bytes32 = bytes32.zeros

    @classmethod
    def from_condition_and_nonce(cls, condition: Program, nonce: bytes32) -> NotarizedPayment:
        with_opcode: Program = Program.to((51, condition))  # Gotta do this because the super class is expecting it
        p = Payment.from_condition(with_opcode)
        puzzle_hash, amount, memos = tuple(p.as_condition_args())
        return cls(puzzle_hash, amount, memos, nonce)


@dataclass(frozen=True, eq=False)
class Offer:
    requested_payments: dict[
        Optional[bytes32], list[NotarizedPayment]
    ]  # The key is the asset id of the asset being requested
    _bundle: WalletSpendBundle
    driver_dict: dict[bytes32, PuzzleInfo]  # asset_id -> asset driver

    # this is a cache of the coin additions made by the SpendBundle (_bundle)
    # ordered by the coin being spent
    _additions: dict[Coin, list[Coin]] = field(init=False, repr=False)
    _hints: dict[bytes32, bytes32] = field(init=False)
    _offered_coins: dict[Optional[bytes32], list[Coin]] = field(init=False, repr=False)
    _final_spend_bundle: Optional[WalletSpendBundle] = field(init=False, repr=False)
    _conditions: Optional[dict[Coin, list[Condition]]] = field(init=False)

    @staticmethod
    def ph() -> bytes32:
        return OFFER_MOD_HASH

    @staticmethod
    def notarize_payments(
        requested_payments: dict[Optional[bytes32], list[Payment]],  # `None` means you are requesting XCK
        coins: list[Coin],
    ) -> dict[Optional[bytes32], list[NotarizedPayment]]:
        # This sort should be reproducible in KLVM with `>s`
        sorted_coins: list[Coin] = sorted(coins, key=Coin.name)
        sorted_coin_list: list[list[Union[bytes32, uint64]]] = [coin_as_list(c) for c in sorted_coins]
        nonce: bytes32 = Program.to(sorted_coin_list).get_tree_hash()

        notarized_payments: dict[Optional[bytes32], list[NotarizedPayment]] = {}
        for asset_id, payments in requested_payments.items():
            notarized_payments[asset_id] = []
            for p in payments:
                puzzle_hash, amount, memos = tuple(p.as_condition_args())
                notarized_payments[asset_id].append(NotarizedPayment(puzzle_hash, amount, memos, nonce))

        return notarized_payments

    # The announcements returned from this function must be asserted in whatever spend bundle is created by the wallet
    @staticmethod
    def calculate_announcements(
        notarized_payments: dict[Optional[bytes32], list[NotarizedPayment]],
        driver_dict: dict[bytes32, PuzzleInfo],
    ) -> list[AssertPuzzleAnnouncement]:
        announcements: list[AssertPuzzleAnnouncement] = []
        for asset_id, payments in notarized_payments.items():
            if asset_id is not None:
                if asset_id not in driver_dict:
                    raise ValueError("Cannot calculate announcements without driver of requested item")
                settlement_ph: bytes32 = construct_puzzle(driver_dict[asset_id], OFFER_MOD).get_tree_hash()
            else:
                settlement_ph = OFFER_MOD_HASH

            msg: bytes32 = Program.to((payments[0].nonce, [p.as_condition_args() for p in payments])).get_tree_hash()
            announcements.append(AssertPuzzleAnnouncement(asserted_ph=settlement_ph, asserted_msg=msg))

        return announcements

    def __post_init__(self) -> None:
        # Verify that there are no duplicate payments
        for payments in self.requested_payments.values():
            payment_programs: list[bytes32] = [p.name() for p in payments]
            if len(set(payment_programs)) != len(payment_programs):
                raise ValueError("Bundle has duplicate requested payments")

        # Verify we have a type for every kind of asset
        for asset_id in self.requested_payments:
            if asset_id is not None and asset_id not in self.driver_dict:
                raise ValueError("Offer does not have enough driver information about the requested payments")

        # populate the _additions cache
        adds: dict[Coin, list[Coin]] = {}
        hints: dict[bytes32, bytes32] = {}
        max_cost = int(DEFAULT_CONSTANTS.MAX_BLOCK_COST_KLVM)
        for cs in self._bundle.coin_spends:
            # you can't spend the same coin twice in the same SpendBundle
            assert cs.coin not in adds
            try:
                hinted_coins, cost = compute_spend_hints_and_additions(cs)
                max_cost -= cost
                adds[cs.coin] = [hc.coin for hc in hinted_coins.values()]
                hints = {**hints, **{id: hc.hint for id, hc in hinted_coins.items() if hc.hint is not None}}
            except Exception:
                continue
            if max_cost < 0:
                raise ValidationError(Err.BLOCK_COST_EXCEEDS_MAX, "compute_additions for CoinSpend")
        object.__setattr__(self, "_additions", adds)
        object.__setattr__(self, "_hints", hints)
        object.__setattr__(self, "_conditions", None)

    def conditions(self) -> dict[Coin, list[Condition]]:
        if self._conditions is None:
            conditions: dict[Coin, list[Condition]] = {}
            max_cost = int(DEFAULT_CONSTANTS.MAX_BLOCK_COST_KLVM)
            for cs in self._bundle.coin_spends:
                try:
                    cost, conds = cs.puzzle_reveal.run_with_cost(max_cost, cs.solution)
                    max_cost -= cost
                    conditions[cs.coin] = parse_conditions_non_consensus(conds.as_iter())
                except Exception:  # pragma: no cover
                    continue
                if max_cost < 0:  # pragma: no cover
                    raise ValidationError(Err.BLOCK_COST_EXCEEDS_MAX, "computing conditions for CoinSpend")
            object.__setattr__(self, "_conditions", conditions)
        assert self._conditions is not None, "self._conditions is None"
        return self._conditions

    def valid_times(self) -> dict[Coin, ConditionValidTimes]:
        return {coin: parse_timelock_info(conditions) for coin, conditions in self.conditions().items()}

    def absolute_valid_times_ban_relatives(self) -> ConditionValidTimes:
        valid_times: ConditionValidTimes = parse_timelock_info(
            [c for conditions in self.conditions().values() for c in conditions]
        )
        if (
            valid_times.max_secs_after_created is not None
            or valid_times.min_secs_since_created is not None
            or valid_times.max_blocks_after_created is not None
            or valid_times.min_blocks_since_created is not None
        ):
            raise ValueError("Offers with relative timelocks are not currently supported")
        return valid_times

    def hints(self) -> dict[bytes32, bytes32]:
        return self._hints

    def additions(self) -> list[Coin]:
        return [c for additions in self._additions.values() for c in additions]

    def removals(self) -> list[Coin]:
        return self._bundle.removals()

    def fees(self) -> int:
        """Unsafe to use for fees validation!!!"""
        amount_in = sum(_.amount for _ in self.removals())
        amount_out = sum(_.amount for _ in self.additions())
        return int(amount_in - amount_out)

    def coin_spends(self) -> list[CoinSpend]:
        return self._bundle.coin_spends

    def aggregated_signature(self) -> G2Element:
        return self._bundle.aggregated_signature

    # This method does not get every coin that is being offered, only the `settlement_payment` children
    # It's also a little heuristic, but it should get most things
    def _get_offered_coins(self) -> dict[Optional[bytes32], list[Coin]]:
        offered_coins: dict[Optional[bytes32], list[Coin]] = {}

        for parent_spend in self._bundle.coin_spends:
            coins_for_this_spend: list[Coin] = []

            parent_puzzle: UncurriedPuzzle = uncurry_puzzle(parent_spend.puzzle_reveal)
            parent_solution: Program = parent_spend.solution.to_program()
            additions: list[Coin] = self._additions[parent_spend.coin]

            puzzle_driver = match_puzzle(parent_puzzle)
            if puzzle_driver is not None:
                asset_id = create_asset_id(puzzle_driver)
                inner_puzzle: Optional[Program] = get_inner_puzzle(puzzle_driver, parent_puzzle)
                inner_solution: Optional[Program] = get_inner_solution(puzzle_driver, parent_solution)
                assert inner_puzzle is not None and inner_solution is not None

                # We're going to look at the conditions created by the inner puzzle
                conditions: Program = inner_puzzle.run(inner_solution)
                expected_num_matches: int = 0
                offered_amounts: list[int] = []
                for condition in conditions.as_iter():
                    if condition.first() == 51 and condition.rest().first() == OFFER_MOD_HASH:
                        expected_num_matches += 1
                        offered_amounts.append(condition.rest().rest().first().as_int())

                # Start by filtering additions that match the amount
                matching_spend_additions = [a for a in additions if a.amount in offered_amounts]

                if len(matching_spend_additions) == expected_num_matches:
                    coins_for_this_spend.extend(matching_spend_additions)
                # We didn't quite get there so now lets narrow it down by puzzle hash
                else:
                    # If we narrowed down too much, we can't trust the amounts so start over with all additions
                    if len(matching_spend_additions) < expected_num_matches:
                        matching_spend_additions = additions
                    matching_spend_additions = [
                        a
                        for a in matching_spend_additions
                        if a.puzzle_hash
                        == construct_puzzle(puzzle_driver, OFFER_MOD_HASH).get_tree_hash_precalc(  # type: ignore
                            OFFER_MOD_HASH
                        )
                    ]
                    if len(matching_spend_additions) == expected_num_matches:
                        coins_for_this_spend.extend(matching_spend_additions)
                    else:
                        raise ValueError("Could not properly guess offered coins from parent spend")
            else:
                # It's much easier if the asset is bare XCK
                asset_id = None
                coins_for_this_spend.extend([a for a in additions if a.puzzle_hash == OFFER_MOD_HASH])

            # We only care about unspent coins
            coins_for_this_spend = [c for c in coins_for_this_spend if c not in self._bundle.removals()]

            if coins_for_this_spend != []:
                offered_coins.setdefault(asset_id, [])
                offered_coins[asset_id].extend(coins_for_this_spend)
        return offered_coins

    def get_offered_coins(self) -> dict[Optional[bytes32], list[Coin]]:
        try:
            if self._offered_coins is not None:
                return self._offered_coins
        except AttributeError:
            object.__setattr__(self, "_offered_coins", self._get_offered_coins())
        return self._offered_coins

    def get_offered_amounts(self) -> dict[Optional[bytes32], int]:
        offered_coins: dict[Optional[bytes32], list[Coin]] = self.get_offered_coins()
        offered_amounts: dict[Optional[bytes32], int] = {}
        for asset_id, coins in offered_coins.items():
            offered_amounts[asset_id] = uint64(sum(c.amount for c in coins))
        return offered_amounts

    def get_requested_payments(self) -> dict[Optional[bytes32], list[NotarizedPayment]]:
        return self.requested_payments

    def get_requested_amounts(self) -> dict[Optional[bytes32], int]:
        requested_amounts: dict[Optional[bytes32], int] = {}
        for asset_id, coins in self.get_requested_payments().items():
            requested_amounts[asset_id] = uint64(sum(c.amount for c in coins))
        return requested_amounts

    def arbitrage(self) -> dict[Optional[bytes32], int]:
        """
        Returns a dictionary of the type of each asset and amount that is involved in the trade
        With the amount being how much their offered amount within the offer
        exceeds/falls short of their requested amount.
        """
        offered_amounts: dict[Optional[bytes32], int] = self.get_offered_amounts()
        requested_amounts: dict[Optional[bytes32], int] = self.get_requested_amounts()

        arbitrage_dict: dict[Optional[bytes32], int] = {}
        for asset_id in [*requested_amounts.keys(), *offered_amounts.keys()]:
            arbitrage_dict[asset_id] = offered_amounts.get(asset_id, 0) - requested_amounts.get(asset_id, 0)

        return arbitrage_dict

    # This is a method mostly for the UI that creates a JSON summary of the offer
    def summary(self) -> tuple[dict[str, int], dict[str, int], dict[str, dict[str, Any]], ConditionValidTimes]:
        offered_amounts: dict[Optional[bytes32], int] = self.get_offered_amounts()
        requested_amounts: dict[Optional[bytes32], int] = self.get_requested_amounts()

        def keys_to_strings(dic: dict[Optional[bytes32], Any]) -> dict[str, Any]:
            new_dic: dict[str, Any] = {}
            for key, val in dic.items():
                if key is None:
                    new_dic["xck"] = val
                else:
                    new_dic[key.hex()] = val
            return new_dic

        driver_dict: dict[str, Any] = {}
        for key, value in self.driver_dict.items():
            driver_dict[key.hex()] = value.info

        return (
            keys_to_strings(offered_amounts),
            keys_to_strings(requested_amounts),
            driver_dict,
            self.absolute_valid_times_ban_relatives(),
        )

    # Also mostly for the UI, returns a dictionary of assets and how much of them is pended for this offer
    # This method is also imperfect for sufficiently complex spends
    def get_pending_amounts(self) -> dict[str, int]:
        all_additions: list[Coin] = self.additions()
        all_removals: list[Coin] = self.removals()
        non_ephemeral_removals: list[Coin] = list(filter(lambda c: c not in all_additions, all_removals))

        pending_dict: dict[str, int] = {}
        # First we add up the amounts of all coins that share an ancestor with the offered coins (i.e. a primary coin)
        for asset_id, coins in self.get_offered_coins().items():
            name = "xck" if asset_id is None else asset_id.hex()
            pending_dict[name] = 0
            for coin in coins:
                root_removal: Coin = self.get_root_removal(coin)

                for addition in filter(lambda c: c.parent_coin_info == root_removal.name(), all_additions):
                    pending_dict[name] += addition.amount

        # Then we gather anything else as unknown
        sum_of_additions_so_far: int = sum(pending_dict.values())
        unknown: int = sum(c.amount for c in non_ephemeral_removals) - sum_of_additions_so_far
        if unknown > 0:
            pending_dict["unknown"] = unknown

        return pending_dict

    # This method returns all of the coins that are being used in the offer (without which it would be invalid)
    def get_involved_coins(self) -> list[Coin]:
        additions = self.additions()
        return list(filter(lambda c: c not in additions, self.removals()))

    # This returns the non-ephemeral removal that is an ancestor of the specified coin
    # This should maybe move to the SpendBundle object at some point
    def get_root_removal(self, coin: Coin) -> Coin:
        all_removals: set[Coin] = set(self.removals())
        all_removal_ids: set[bytes32] = {c.name() for c in all_removals}
        non_ephemeral_removals: set[Coin] = {
            c for c in all_removals if c.parent_coin_info not in {r.name() for r in all_removals}
        }
        if coin.name() not in all_removal_ids and coin.parent_coin_info not in all_removal_ids:
            raise ValueError("The specified coin is not a coin in this bundle")

        while coin not in non_ephemeral_removals:
            coin = next(c for c in all_removals if c.name() == coin.parent_coin_info)

        return coin

    # This will only return coins that are ancestors of settlement payments
    def get_primary_coins(self) -> list[Coin]:
        primary_coins: set[Coin] = set()
        for _, coins in self.get_offered_coins().items():
            for coin in coins:
                primary_coins.add(self.get_root_removal(coin))
        return list(primary_coins)

    # This returns the minimum coins that when spent will invalidate the rest of the bundle
    def get_cancellation_coins(self) -> list[Coin]:
        # First, we're going to gather:
        dependencies: dict[bytes32, list[bytes32]] = {}  # all of the hashes that each coin depends on
        announcements: dict[bytes32, list[bytes32]] = {}  # all of the hashes of the announcement that each coin makes
        coin_names: list[bytes32] = []  # The names of all the coins
        additions = self.additions()
        for spend in [cs for cs in self._bundle.coin_spends if cs.coin not in additions]:
            name = bytes32(spend.coin.name())
            coin_names.append(name)
            dependencies[name] = []
            announcements[name] = []
            conditions: Program = spend.puzzle_reveal.run_with_cost(INFINITE_COST, spend.solution)[1]
            for condition in conditions.as_iter():
                if condition.first() == 60:  # create coin announcement
                    announcements[name].append(
                        AssertCoinAnnouncement(asserted_id=name, asserted_msg=condition.at("rf").as_python()).msg_calc
                    )
                elif condition.first() == 61:  # assert coin announcement
                    dependencies[name].append(bytes32(condition.at("rf").as_python()))

        # We now enter a loop that is attempting to express the following logic:
        # "If I am depending on another coin in the same bundle, you may as well cancel that coin instead of me"
        # By the end of the loop, we should have filtered down the list of coin_names to include only those that will
        # cancel everything else
        while True:
            removed = detect_dependent_coin(coin_names, dependencies, announcements)
            if removed is None:
                break
            removed_coin, provider = removed
            removed_announcements: list[bytes32] = announcements[removed_coin]
            remove_these_keys: list[bytes32] = [removed_coin]
            while True:
                for coin, deps in dependencies.items():
                    if set(deps) & set(removed_announcements) and coin != provider:
                        remove_these_keys.append(coin)
                removed_announcements = []
                for coin in remove_these_keys:
                    dependencies.pop(coin)
                    removed_announcements.extend(announcements.pop(coin))
                coin_names = [n for n in coin_names if n not in remove_these_keys]
                if removed_announcements == []:
                    break
                else:
                    remove_these_keys = []

        return [cs.coin for cs in self._bundle.coin_spends if cs.coin.name() in coin_names]

    @classmethod
    def aggregate(cls, offers: list[Offer]) -> Offer:
        total_requested_payments: dict[Optional[bytes32], list[NotarizedPayment]] = {}
        total_bundle = WalletSpendBundle([], G2Element())
        total_driver_dict: dict[bytes32, PuzzleInfo] = {}
        for offer in offers:
            # First check for any overlap in inputs
            total_inputs: set[Coin] = {cs.coin for cs in total_bundle.coin_spends}
            offer_inputs: set[Coin] = {cs.coin for cs in offer._bundle.coin_spends}
            if total_inputs & offer_inputs:
                raise ValueError("The aggregated offers overlap inputs")

            # Next, do the aggregation
            for asset_id, payments in offer.requested_payments.items():
                if asset_id in total_requested_payments:
                    total_requested_payments[asset_id].extend(payments)
                else:
                    total_requested_payments[asset_id] = payments

            for key, value in offer.driver_dict.items():
                if key in total_driver_dict and total_driver_dict[key] != value:
                    raise ValueError(f"The offers to aggregate disagree on the drivers for {key.hex()}")

            total_bundle = WalletSpendBundle.aggregate([total_bundle, offer._bundle])
            total_driver_dict.update(offer.driver_dict)

        return cls(total_requested_payments, total_bundle, total_driver_dict)

    # Validity is defined by having enough funds within the offer to satisfy both sides
    def is_valid(self) -> bool:
        return all([value >= 0 for value in self.arbitrage().values()])

    # A "valid" spend means that this bundle can be pushed to the network and will succeed
    # This differs from the `to_spend_bundle` method which deliberately creates an invalid SpendBundle
    def to_valid_spend(self, arbitrage_ph: Optional[bytes32] = None, solver: Solver = Solver({})) -> WalletSpendBundle:
        if not self.is_valid():
            raise ValueError("Offer is currently incomplete")

        completion_spends: list[CoinSpend] = []
        all_offered_coins: dict[Optional[bytes32], list[Coin]] = self.get_offered_coins()
        total_arbitrage_amount: dict[Optional[bytes32], int] = self.arbitrage()
        for asset_id, payments in self.requested_payments.items():
            offered_coins: list[Coin] = all_offered_coins[asset_id]

            # Because of CAT supply laws, we must specify a place for the leftovers to go
            arbitrage_amount: int = total_arbitrage_amount[asset_id]
            all_payments: list[NotarizedPayment] = payments.copy()
            if arbitrage_amount > 0:
                assert arbitrage_amount is not None
                assert arbitrage_ph is not None
                all_payments.append(NotarizedPayment(arbitrage_ph, uint64(arbitrage_amount)))

            # Some assets need to know about siblings so we need to collect all spends first to be able to use them
            coin_to_spend_dict: dict[Coin, CoinSpend] = {}
            coin_to_solution_dict: dict[Coin, Program] = {}
            for coin in offered_coins:
                parent_spend: CoinSpend = next(
                    filter(lambda cs: cs.coin.name() == coin.parent_coin_info, self._bundle.coin_spends)
                )
                coin_to_spend_dict[coin] = parent_spend

                inner_solutions = []
                if coin == offered_coins[0]:
                    nonces: list[bytes32] = [p.nonce for p in all_payments]
                    for nonce in list(dict.fromkeys(nonces)):  # dedup without messing with order
                        nonce_payments: list[NotarizedPayment] = list(filter(lambda p: p.nonce == nonce, all_payments))
                        inner_solutions.append((nonce, [np.as_condition_args() for np in nonce_payments]))
                coin_to_solution_dict[coin] = Program.to(inner_solutions)

            for coin in offered_coins:
                if asset_id:
                    siblings: str = "("
                    sibling_spends: str = "("
                    sibling_puzzles: str = "("
                    sibling_solutions: str = "("
                    disassembled_offer_mod: str = disassemble(OFFER_MOD)
                    for sibling_coin in offered_coins:
                        if sibling_coin != coin:
                            siblings += (
                                "0x"
                                + sibling_coin.parent_coin_info.hex()
                                + sibling_coin.puzzle_hash.hex()
                                + uint64(sibling_coin.amount).stream_to_bytes().hex()
                                + " "
                            )
                            sibling_spends += "0x" + bytes(coin_to_spend_dict[sibling_coin]).hex() + " "
                            sibling_puzzles += disassembled_offer_mod + " "
                            sibling_solutions += disassemble(coin_to_solution_dict[sibling_coin]) + " "
                    siblings += ")"
                    sibling_spends += ")"
                    sibling_puzzles += ")"
                    sibling_solutions += ")"

                    solution: Program = solve_puzzle(
                        self.driver_dict[asset_id],
                        Solver(
                            {
                                "coin": "0x"
                                + coin.parent_coin_info.hex()
                                + coin.puzzle_hash.hex()
                                + uint64(coin.amount).stream_to_bytes().hex(),
                                "parent_spend": "0x" + bytes(coin_to_spend_dict[coin]).hex(),
                                "siblings": siblings,
                                "sibling_spends": sibling_spends,
                                "sibling_puzzles": sibling_puzzles,
                                "sibling_solutions": sibling_solutions,
                                **solver.info,
                            }
                        ),
                        OFFER_MOD,
                        Program.to(coin_to_solution_dict[coin]),
                    )
                else:
                    solution = Program.to(coin_to_solution_dict[coin])

                completion_spends.append(
                    make_spend(
                        coin,
                        construct_puzzle(self.driver_dict[asset_id], OFFER_MOD) if asset_id else OFFER_MOD,
                        solution,
                    )
                )

        return WalletSpendBundle.aggregate([WalletSpendBundle(completion_spends, G2Element()), self._bundle])

    def to_spend_bundle(self) -> WalletSpendBundle:
        try:
            if self._final_spend_bundle is not None:
                return self._final_spend_bundle
        except AttributeError:
            pass
        # Before we serialize this as a SpendBundle, we need to serialize the `requested_payments` as dummy CoinSpends
        additional_coin_spends: list[CoinSpend] = []
        for asset_id, payments in self.requested_payments.items():
            puzzle_reveal: Program = construct_puzzle(self.driver_dict[asset_id], OFFER_MOD) if asset_id else OFFER_MOD
            inner_solutions = []
            nonces: list[bytes32] = [p.nonce for p in payments]
            for nonce in list(dict.fromkeys(nonces)):  # dedup without messing with order
                nonce_payments: list[NotarizedPayment] = list(filter(lambda p: p.nonce == nonce, payments))
                inner_solutions.append((nonce, [np.as_condition_args() for np in nonce_payments]))

            additional_coin_spends.append(
                make_spend(
                    Coin(
                        bytes32.zeros,
                        puzzle_reveal.get_tree_hash(),
                        uint64(0),
                    ),
                    puzzle_reveal,
                    Program.to(inner_solutions),
                )
            )

        sb = WalletSpendBundle.aggregate(
            [
                WalletSpendBundle(additional_coin_spends, G2Element()),
                self._bundle,
            ]
        )
        object.__setattr__(self, "_final_spend_bundle", sb)
        return sb

    @classmethod
    def from_spend_bundle(cls, bundle: WalletSpendBundle) -> Offer:
        # Because of the `to_spend_bundle` method, we need to parse the dummy CoinSpends as `requested_payments`
        requested_payments: dict[Optional[bytes32], list[NotarizedPayment]] = {}
        driver_dict: dict[bytes32, PuzzleInfo] = {}
        leftover_coin_spends: list[CoinSpend] = []
        for coin_spend in bundle.coin_spends:
            driver = match_puzzle(uncurry_puzzle(coin_spend.puzzle_reveal))
            if driver is not None:
                asset_id = create_asset_id(driver)
                assert asset_id is not None
                driver_dict[asset_id] = driver
            else:
                asset_id = None
            if coin_spend.coin.parent_coin_info == bytes32.zeros:
                notarized_payments: list[NotarizedPayment] = []
                for payment_group in coin_spend.solution.to_program().as_iter():
                    nonce = bytes32(payment_group.first().as_atom())
                    payment_args_list = payment_group.rest().as_iter()
                    notarized_payments.extend(
                        [NotarizedPayment.from_condition_and_nonce(condition, nonce) for condition in payment_args_list]
                    )

                requested_payments[asset_id] = notarized_payments
            else:
                leftover_coin_spends.append(coin_spend)

        return cls(
            requested_payments, WalletSpendBundle(leftover_coin_spends, bundle.aggregated_signature), driver_dict
        )

    def name(self) -> bytes32:
        return self.to_spend_bundle().name()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Offer):
            return False  # don't attempt to compare against unrelated types
        return self.name() == other.name()

    def compress(self, version: Optional[int] = None) -> bytes:
        as_spend_bundle = self.to_spend_bundle()
        if version is None:
            mods: list[bytes] = [bytes(s.puzzle_reveal.to_program().uncurry()[0]) for s in as_spend_bundle.coin_spends]
            version = max(lowest_best_version(mods), 6)  # Clients lower than version 6 should not be able to parse
        return compress_object_with_puzzles(bytes(as_spend_bundle), version)

    @classmethod
    def from_compressed(cls, compressed_bytes: bytes) -> Offer:
        return Offer.from_bytes(decompress_object_with_puzzles(compressed_bytes))

    @classmethod
    def try_offer_decompression(cls, offer_bytes: bytes) -> Offer:
        try:
            return cls.from_compressed(offer_bytes)
        except TypeError:
            pass
        return cls.from_bytes(offer_bytes)

    def to_bech32(self, prefix: str = "offer", compression_version: Optional[int] = None) -> str:
        offer_bytes = self.compress(version=compression_version)
        encoded = bech32_encode(prefix, convertbits(list(offer_bytes), 8, 5))
        return encoded

    @classmethod
    def from_bech32(cls, offer_bech32: str) -> Offer:
        _hrpgot, data = bech32_decode(offer_bech32, max_length=len(offer_bech32))
        if data is None:
            raise ValueError("Invalid Offer")
        decoded = convertbits(list(data), 5, 8, False)
        decoded_bytes = bytes(decoded)
        return cls.try_offer_decompression(decoded_bytes)

    # Methods to make this a valid Streamable member
    # We basically hijack the SpendBundle versions for most of it
    @classmethod
    def parse(cls, f: BinaryIO) -> Offer:
        parsed_bundle = parse_rust(f, WalletSpendBundle)
        return cls.from_bytes(bytes(parsed_bundle))

    def stream(self, f: BinaryIO) -> None:
        spend_bundle_bytes = self.to_spend_bundle().to_bytes()
        f.write(spend_bundle_bytes)

    def __bytes__(self) -> bytes:
        return bytes(self.to_spend_bundle())

    @classmethod
    def from_bytes(cls, as_bytes: bytes) -> Offer:
        # Because of the __bytes__ method, we need to parse the dummy CoinSpends as `requested_payments`
        bundle = WalletSpendBundle.from_bytes(as_bytes)
        return cls.from_spend_bundle(bundle)



from __future__ import annotations

import random
import time

from chik_rs import AugSchemeMPL, PrivateKey
from klvm_tools import binutils

from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.simulator.wallet_tools import WalletTool
from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.condition_with_args import ConditionWithArgs
from chik.util.ints import uint32
from chik.wallet.derive_keys import master_sk_to_wallet_sk
from chik.wallet.puzzles.p2_delegated_puzzle import puzzle_for_pk


def float_to_str(f):
    float_string = repr(f)
    if "e" in float_string:  # detect scientific notation
        digits, exp_str = float_string.split("e")
        digits = digits.replace(".", "").replace("-", "")
        exp = int(exp_str)
        zero_padding = "0" * (abs(int(exp)) - 1)  # minus 1 for decimal point in the sci notation
        sign = "-" if f < 0 else ""
        if exp > 0:
            float_string = f"{sign}{digits}{zero_padding}.0"
        else:
            float_string = f"{sign}0.{zero_padding}{digits}"
    return float_string


def run_and_return_cost_time(chiklisp):
    start = time.time()
    klvm_loop = "((c (q ((c (f (a)) (c (f (a)) (c (f (r (a))) (c (f (r (r (a))))"
    " (q ()))))))) (c (q ((c (i (f (r (a))) (q (i (q 1) ((c (f (a)) (c (f (a))"
    " (c (- (f (r (a))) (q 1)) (c (f (r (r (a)))) (q ()))))))"
    " ((c (f (r (r (a)))) (q ()))))) (q (q ()))) (a)))) (a))))"
    loop_program = Program.to(binutils.assemble(klvm_loop))
    klvm_loop_solution = f"(1000 {chiklisp})"
    solution_program = Program.to(binutils.assemble(klvm_loop_solution))

    cost, _ = loop_program.run_with_cost(INFINITE_COST, solution_program)

    end = time.time()
    total_time = end - start

    return cost, total_time


def get_cost_compared_to_addition(addition_cost, addition_time, other_time):
    return (addition_cost * other_time) / addition_time


def benchmark_all_operators():
    addition = "(+ (q 1000000000) (q 1000000000))"
    substraction = "(- (q 1000000000) (q 1000000000))"
    multiply = "(* (q 1000000000) (q 1000000000))"
    greater = "(> (q 1000000000) (q 1000000000))"
    equal = "(= (q 1000000000) (q 1000000000))"
    if_klvm = "(i (= (q 1000000000) (q 1000000000)) (q 1000000000) (q 1000000000))"
    sha256tree = "(sha256 (q 1000000000))"
    pubkey_for_exp = "(pubkey_for_exp (q 1))"
    point_add = "(point_add"
    " (q 0x17f1d3a73197d7942695638c4fa9ac0fc3688c4f9774b905a14e3a3f171bac586c55e83ff97a1aeffb3af00adb22c6bb)"
    " (q 0x17f1d3a73197d7942695638c4fa9ac0fc3688c4f9774b905a14e3a3f171bac586c55e83ff97a1aeffb3af00adb22c6bb))"
    _point_add_cost, point_add_time = run_and_return_cost_time(point_add)
    addition_cost, addition_time = run_and_return_cost_time(addition)
    _substraction_cost, substraction_time = run_and_return_cost_time(substraction)
    _multiply_cost, multiply_time = run_and_return_cost_time(multiply)
    _greater_cost, greater_time = run_and_return_cost_time(greater)
    _equal_cost, equal_time = run_and_return_cost_time(equal)
    _if_cost, if_time = run_and_return_cost_time(if_klvm)
    _sha256tree_cost, sha256tree_time = run_and_return_cost_time(sha256tree)
    _pubkey_for_exp_cost, pubkey_for_exp_time = run_and_return_cost_time(pubkey_for_exp)

    one_addition = 1
    one_substraction = get_cost_compared_to_addition(addition_cost, addition_time, substraction_time) / addition_cost
    one_multiply = get_cost_compared_to_addition(addition_cost, addition_time, multiply_time) / addition_cost
    one_greater = get_cost_compared_to_addition(addition_cost, addition_time, greater_time) / addition_cost
    one_equal = get_cost_compared_to_addition(addition_cost, addition_time, equal_time) / addition_cost
    one_if = get_cost_compared_to_addition(addition_cost, addition_time, if_time) / addition_cost
    one_sha256 = get_cost_compared_to_addition(addition_cost, addition_time, sha256tree_time) / addition_cost
    one_pubkey_for_exp = (
        get_cost_compared_to_addition(addition_cost, addition_time, pubkey_for_exp_time) / addition_cost
    )
    one_point_add = get_cost_compared_to_addition(addition_cost, addition_time, point_add_time) / addition_cost

    print(f"cost of addition is: {one_addition}")
    print(f"cost of one_substraction is: {one_substraction}")
    print(f"cost of one_multiply is: {one_multiply}")
    print(f"cost of one_greater is: {one_greater}")
    print(f"cost of one_equal is: {one_equal}")
    print(f"cost of one_if is: {one_if}")
    print(f"cost of one_sha256 is: {one_sha256}")
    print(f"cost of one_pubkey_for_exp is: {one_pubkey_for_exp}")
    print(f"cost of one_point_add is: {one_point_add}")


if __name__ == "__main__":
    """
    Naive way to calculate cost ratio between vByte and KLVM cost unit.
    AggSig has assigned cost of 20vBytes, simple KLVM program is benchmarked against it.
    """
    wallet_tool = WalletTool(DEFAULT_CONSTANTS)
    benchmark_all_operators()
    secret_key: PrivateKey = AugSchemeMPL.key_gen(bytes([2] * 32))
    puzzles = []
    solutions = []
    private_keys = []
    public_keys = []

    seeded_random = random.Random()
    seeded_random.seed(a=0, version=2)

    for i in range(0, 1000):
        private_key: PrivateKey = master_sk_to_wallet_sk(secret_key, uint32(i))
        public_key = private_key.get_g1()
        solution = wallet_tool.make_solution(
            {
                ConditionOpcode.ASSERT_MY_COIN_ID: [
                    ConditionWithArgs(ConditionOpcode.ASSERT_MY_COIN_ID, [bytes32.random(seeded_random)])
                ]
            }
        )
        puzzle = puzzle_for_pk(bytes(public_key))
        puzzles.append(puzzle)
        solutions.append(solution)
        private_keys.append(private_key)
        public_keys.append(public_key)

    # Run Puzzle 1000 times
    puzzle_start = time.time()
    klvm_cost = 0
    for i in range(0, 1000):
        cost_run, _ = puzzles[i].run_with_cost(INFINITE_COST, solutions[i])
        klvm_cost += cost_run

    puzzle_end = time.time()
    puzzle_time = puzzle_end - puzzle_start
    print(f"Puzzle_time is: {puzzle_time}")
    print(f"Puzzle cost sum is: {klvm_cost}")

    private_key = master_sk_to_wallet_sk(secret_key, uint32(0))
    public_key = private_key.get_g1()
    message = bytes32.random(seeded_random)
    signature = AugSchemeMPL.sign(private_key, message)
    pk_message_pair = (public_key, message)

    # Run AggSig 1000 times
    agg_sig_start = time.time()
    agg_sig_cost = 0
    for i in range(0, 1000):
        valid = AugSchemeMPL.verify(public_key, message, signature)
        assert valid
        agg_sig_cost += 20
    agg_sig_end = time.time()
    agg_sig_time = agg_sig_end - agg_sig_start
    print(f"Aggsig Cost: {agg_sig_cost}")
    print(f"Aggsig time is: {agg_sig_time}")

    # klvm_should_cost = agg_sig_cost * puzzle_time / agg_sig_time
    klvm_should_cost = (agg_sig_cost * puzzle_time) / agg_sig_time
    print(f"Puzzle should cost: {klvm_should_cost}")
    constant = klvm_should_cost / klvm_cost
    format = float_to_str(constant)
    print(f"Constant factor: {format}")
    print(f"KLVM RATIO MULTIPLIER: {1 / constant}")



from __future__ import annotations

from dataclasses import dataclass

from chik.types.blockchain_format.coin import Coin as _Coin
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import CoinSpend
from chik.util.ints import uint64
from chik.util.streamable import Streamable
from chik.wallet.util.klvm_streamable import klvm_streamable

# This file contains the base types for communication between a wallet and an offline transaction signer.
# These types should be compliant with CHIP-0028


@klvm_streamable
@dataclass(frozen=True)
class Coin(Streamable):
    parent_coin_id: bytes32
    puzzle_hash: bytes32
    amount: uint64


@klvm_streamable
@dataclass(frozen=True)
class Spend(Streamable):
    coin: Coin
    puzzle: Program
    solution: Program

    @classmethod
    def from_coin_spend(cls, coin_spend: CoinSpend) -> Spend:
        return cls(
            Coin(
                coin_spend.coin.parent_coin_info,
                coin_spend.coin.puzzle_hash,
                uint64(coin_spend.coin.amount),
            ),
            coin_spend.puzzle_reveal.to_program(),
            coin_spend.solution.to_program(),
        )

    def as_coin_spend(self) -> CoinSpend:
        return CoinSpend(
            _Coin(
                self.coin.parent_coin_id,
                self.coin.puzzle_hash,
                self.coin.amount,
            ),
            SerializedProgram.from_program(self.puzzle),
            SerializedProgram.from_program(self.solution),
        )


@klvm_streamable
@dataclass(frozen=True)
class TransactionInfo(Streamable):
    spends: list[Spend]


@klvm_streamable
@dataclass(frozen=True)
class SigningTarget(Streamable):
    fingerprint: bytes
    message: bytes
    hook: bytes32


@klvm_streamable
@dataclass(frozen=True)
class SumHint(Streamable):
    fingerprints: list[bytes]
    synthetic_offset: bytes
    final_pubkey: bytes


@klvm_streamable
@dataclass(frozen=True)
class PathHint(Streamable):
    root_fingerprint: bytes
    path: list[uint64]


@klvm_streamable
@dataclass(frozen=True)
class KeyHints(Streamable):
    sum_hints: list[SumHint]
    path_hints: list[PathHint]


@klvm_streamable
@dataclass(frozen=True)
class SigningInstructions(Streamable):
    key_hints: KeyHints
    targets: list[SigningTarget]


@klvm_streamable
@dataclass(frozen=True)
class UnsignedTransaction(Streamable):
    transaction_info: TransactionInfo
    signing_instructions: SigningInstructions


@klvm_streamable
@dataclass(frozen=True)
class SigningResponse(Streamable):
    signature: bytes
    hook: bytes32


@klvm_streamable
@dataclass(frozen=True)
class Signature(Streamable):
    type: str
    signature: bytes


@klvm_streamable
@dataclass(frozen=True)
class SignedTransaction(Streamable):
    transaction_info: TransactionInfo
    signatures: list[Signature]



from __future__ import annotations

import dataclasses
from random import Random

import pytest

# TODO: update after resolution in https://github.com/pytest-dev/pytest/issues/7469
from _pytest.fixtures import SubRequest

from chik._tests.util.misc import Marks, datacases, measure_runtime
from chik.data_layer.data_layer_util import (
    ClearPendingRootsRequest,
    ClearPendingRootsResponse,
    ProofOfInclusion,
    ProofOfInclusionLayer,
    Root,
    Side,
    Status,
    internal_hash,
    key_hash,
    leaf_hash,
)
from chik.rpc.data_layer_rpc_util import MarshallableProtocol
from chik.types.blockchain_format.program import Program
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32

pytestmark = pytest.mark.data_layer


def create_valid_proof_of_inclusion(layer_count: int, other_hash_side: Side) -> ProofOfInclusion:
    node_hash = bytes32(b"a" * 32)
    layers: list[ProofOfInclusionLayer] = []

    existing_hash = node_hash

    other_hashes = [bytes32([i] * 32) for i in range(layer_count)]

    for other_hash in other_hashes:
        new_layer = ProofOfInclusionLayer.from_hashes(
            primary_hash=existing_hash,
            other_hash_side=other_hash_side,
            other_hash=other_hash,
        )

        layers.append(new_layer)
        existing_hash = new_layer.combined_hash

    return ProofOfInclusion(node_hash=node_hash, layers=layers)


@pytest.fixture(name="side", params=[Side.LEFT, Side.RIGHT])
def side_fixture(request: SubRequest) -> Side:
    # https://github.com/pytest-dev/pytest/issues/8763
    return request.param  # type: ignore[no-any-return]


@pytest.fixture(name="valid_proof_of_inclusion", params=[0, 1, 5])
def valid_proof_of_inclusion_fixture(request: SubRequest, side: Side) -> ProofOfInclusion:
    return create_valid_proof_of_inclusion(layer_count=request.param, other_hash_side=side)


@pytest.fixture(
    name="invalid_proof_of_inclusion",
    params=["bad root hash", "bad other hash", "bad other side", "bad node hash"],
)
def invalid_proof_of_inclusion_fixture(request: SubRequest, side: Side) -> ProofOfInclusion:
    valid_proof_of_inclusion = create_valid_proof_of_inclusion(layer_count=5, other_hash_side=side)

    layers = list(valid_proof_of_inclusion.layers)
    a_hash = bytes32(b"f" * 32)

    if request.param == "bad root hash":
        layers[-1] = dataclasses.replace(layers[-1], combined_hash=a_hash)
        return dataclasses.replace(valid_proof_of_inclusion, layers=layers)
    elif request.param == "bad other hash":
        layers[1] = dataclasses.replace(layers[1], other_hash=a_hash)
        return dataclasses.replace(valid_proof_of_inclusion, layers=layers)
    elif request.param == "bad other side":
        layers[1] = dataclasses.replace(layers[1], other_hash_side=layers[1].other_hash_side.other())
        return dataclasses.replace(valid_proof_of_inclusion, layers=layers)
    elif request.param == "bad node hash":
        return dataclasses.replace(valid_proof_of_inclusion, node_hash=a_hash)

    raise Exception(f"Unhandled parametrization: {request.param!r}")  # pragma: no cover


def test_proof_of_inclusion_is_valid(valid_proof_of_inclusion: ProofOfInclusion) -> None:
    assert valid_proof_of_inclusion.valid()


def test_proof_of_inclusion_is_invalid(invalid_proof_of_inclusion: ProofOfInclusion) -> None:
    assert not invalid_proof_of_inclusion.valid()


@dataclasses.dataclass()
class RoundTripCase:
    id: str
    instance: MarshallableProtocol
    marks: Marks = ()


@datacases(
    RoundTripCase(
        id="Root",
        instance=Root(
            store_id=bytes32.zeros,
            node_hash=bytes32(b"\x01" * 32),
            generation=3,
            status=Status.PENDING,
        ),
    ),
    RoundTripCase(
        id="ClearPendingRootsRequest",
        instance=ClearPendingRootsRequest(store_id=bytes32(b"\x12" * 32)),
    ),
    RoundTripCase(
        id="ClearPendingRootsResponse success",
        instance=ClearPendingRootsResponse(
            success=True,
            root=Root(
                store_id=bytes32.zeros,
                node_hash=bytes32(b"\x01" * 32),
                generation=3,
                status=Status.PENDING,
            ),
        ),
    ),
    RoundTripCase(
        id="ClearPendingRootsResponse failure",
        instance=ClearPendingRootsResponse(success=False, root=None),
    ),
)
def test_marshalling_round_trip(case: RoundTripCase) -> None:
    marshalled = case.instance.marshal()
    unmarshalled = type(case.instance).unmarshal(marshalled)
    assert case.instance == unmarshalled


def test_internal_hash(seeded_random: Random) -> None:
    def definition(left_hash: bytes32, right_hash: bytes32) -> bytes32:
        return Program.to((left_hash, right_hash)).get_tree_hash_precalc(left_hash, right_hash)

    data: list[tuple[bytes32, bytes32, bytes32]] = []
    for _ in range(5000):
        left_hash = bytes32.random(r=seeded_random)
        right_hash = bytes32.random(r=seeded_random)
        reference = definition(left_hash=left_hash, right_hash=right_hash)
        data.append((left_hash, right_hash, reference))

    with measure_runtime(label="optimization"):
        for left_hash, right_hash, reference in data:
            assert internal_hash(left_hash=left_hash, right_hash=right_hash) == reference

    with measure_runtime(label="definition"):
        for left_hash, right_hash, reference in data:
            assert definition(left_hash=left_hash, right_hash=right_hash) == reference


def get_random_bytes(length: int, r: Random) -> bytes:
    if length == 0:
        return b""

    return r.getrandbits(length * 8).to_bytes(length, "big")


def test_leaf_hash(seeded_random: Random) -> None:
    def definition(key: bytes, value: bytes) -> bytes32:
        return SerializedProgram.to((key, value)).get_tree_hash()

    data: list[tuple[bytes, bytes, bytes32]] = []
    for cycle in range(20000):
        if cycle in {0, 1}:
            length = 0
        else:
            length = seeded_random.randrange(100)

        key = get_random_bytes(length=length, r=seeded_random)

        if cycle in {1, 2}:
            length = 0
        else:
            length = seeded_random.randrange(100)
        value = get_random_bytes(length=length, r=seeded_random)
        reference = definition(key=key, value=value)
        data.append((key, value, reference))

    with measure_runtime(label="optimization"):
        for key, value, reference in data:
            assert leaf_hash(key=key, value=value) == reference

    with measure_runtime(label="definition"):
        for key, value, reference in data:
            assert definition(key=key, value=value) == reference


def test_key_hash(seeded_random: Random) -> None:
    def definition(key: bytes) -> bytes32:
        return SerializedProgram.to(key).get_tree_hash()

    data: list[tuple[bytes, bytes32]] = []
    for cycle in range(30000):
        if cycle == 0:
            length = 0
        else:
            length = seeded_random.randrange(100)
        key = get_random_bytes(length=length, r=seeded_random)
        reference = definition(key=key)
        data.append((key, reference))

    with measure_runtime(label="optimization"):
        for key, reference in data:
            assert key_hash(key=key) == reference

    with measure_runtime(label="definition"):
        for key, reference in data:
            assert definition(key=key) == reference



from __future__ import annotations

import os
import shutil
import tempfile
from pathlib import Path
from typing import Optional
from unittest.mock import patch

from chik.util.file_keyring import FileKeyring, keyring_path_from_root
from chik.util.keychain import Keychain
from chik.util.keyring_wrapper import KeyringWrapper


def setup_mock_file_keyring(mock_configure_backend, temp_file_keyring_dir, populate=False):
    if populate:
        # Populate the file keyring with an empty (but encrypted) data set
        file_keyring_path = keyring_path_from_root(Path(temp_file_keyring_dir))
        os.makedirs(os.path.dirname(file_keyring_path), 0o700, True)
        with open(
            os.open(
                keyring_path_from_root(Path(temp_file_keyring_dir)),
                os.O_CREAT | os.O_WRONLY | os.O_TRUNC,
                0o600,
            ),
            "w",
        ) as f:
            f.write(
                # Encrypted using DEFAULT_PASSPHRASE_IF_NO_MASTER_PASSPHRASE. Data holds an empty keyring.
                "data: xtcxYOWtbeO9ruv4Nkwhw1pcTJCNh/fvPSdFxez/L0ysnag=\n"
                "nonce: 17ecac58deb7a392fccef49e\n"
                "salt: b1aa32d5730288d653e82017e4a4057c\n"
                "version: 1"
            )

    # Create the file keyring
    mock_configure_backend.return_value = FileKeyring.create(keys_root_path=Path(temp_file_keyring_dir))


class TempKeyring:
    def __init__(
        self,
        *,
        user: str = "testing-1.8.0",
        service: str = "testing-chik-1.8.0",
        populate: bool = False,
        existing_keyring_path: Optional[str] = None,
        delete_on_cleanup: bool = True,
        use_os_credential_store: bool = False,
    ):
        self.keychain = self._patch_and_create_keychain(
            user=user,
            service=service,
            populate=populate,
            existing_keyring_path=existing_keyring_path,
            use_os_credential_store=use_os_credential_store,
        )
        self.old_keys_root_path = None
        self.delete_on_cleanup = delete_on_cleanup
        self.cleaned_up = False

    def _patch_and_create_keychain(
        self,
        *,
        user: str,
        service: str,
        populate: bool,
        existing_keyring_path: Optional[str],
        use_os_credential_store: bool,
    ):
        existing_keyring_dir = Path(existing_keyring_path).parent if existing_keyring_path else None
        temp_dir = existing_keyring_dir or tempfile.mkdtemp(prefix="test_keyring_wrapper")

        mock_supports_os_passphrase_storage_patch = patch("chik.util.keychain.supports_os_passphrase_storage")
        mock_supports_os_passphrase_storage = mock_supports_os_passphrase_storage_patch.start()

        # Patch supports_os_passphrase_storage() to return use_os_credential_store
        mock_supports_os_passphrase_storage.return_value = use_os_credential_store

        mock_configure_backend_patch = patch.object(KeyringWrapper, "_configure_backend")
        mock_configure_backend = mock_configure_backend_patch.start()
        setup_mock_file_keyring(mock_configure_backend, temp_dir, populate=populate)

        keychain = Keychain(user=user, service=service)
        keychain.keyring_wrapper = KeyringWrapper(keys_root_path=Path(temp_dir))

        # Stash the temp_dir in the keychain instance
        keychain._temp_dir = temp_dir  # type: ignore

        # Stash the patches in the keychain instance
        keychain._mock_supports_os_passphrase_storage_patch = mock_supports_os_passphrase_storage_patch  # type: ignore
        keychain._mock_configure_backend_patch = mock_configure_backend_patch  # type: ignore

        return keychain

    def __enter__(self):
        assert not self.cleaned_up
        if KeyringWrapper.get_shared_instance(create_if_necessary=False) is not None:
            self.old_keys_root_path = KeyringWrapper.get_shared_instance().keys_root_path
            KeyringWrapper.cleanup_shared_instance()
        kc = self.get_keychain()
        KeyringWrapper.set_keys_root_path(kc.keyring_wrapper.keys_root_path)
        return kc

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.cleanup()

    def get_keychain(self) -> Keychain:
        return self.keychain

    def cleanup(self) -> None:
        assert not self.cleaned_up

        keys_root_path = self.keychain.keyring_wrapper.keys_root_path

        if self.delete_on_cleanup:
            self.keychain.keyring_wrapper.keyring.cleanup_keyring_file_watcher()
            shutil.rmtree(self.keychain._temp_dir)

        if self.old_keys_root_path is not None:
            if KeyringWrapper.get_shared_instance(create_if_necessary=False) is not None:
                shared_keys_root_path = KeyringWrapper.get_shared_instance().keys_root_path
                if shared_keys_root_path == keys_root_path:
                    KeyringWrapper.cleanup_shared_instance()
                    KeyringWrapper.set_keys_root_path(self.old_keys_root_path)
                    KeyringWrapper.get_shared_instance()

        self.cleaned_up = True



from __future__ import annotations

import asyncio
import os
import sys
import tempfile
import time
from pathlib import Path
from typing import Optional

from chik.data_layer.data_layer_util import Side, TerminalNode, leaf_hash
from chik.data_layer.data_store import DataStore
from chik.types.blockchain_format.sized_bytes import bytes32


async def generate_datastore(num_nodes: int, slow_mode: bool) -> None:
    with tempfile.TemporaryDirectory() as temp_directory:
        temp_directory_path = Path(temp_directory)
        db_path = temp_directory_path.joinpath("dl_benchmark.sqlite")
        print(f"Writing DB to {db_path}")

        if os.path.exists(db_path):
            os.remove(db_path)

        async with DataStore.managed(database=db_path) as data_store:
            store_id = bytes32(b"0" * 32)
            await data_store.create_tree(store_id)

            insert_time = 0.0
            insert_count = 0
            autoinsert_time = 0.0
            autoinsert_count = 0
            delete_time = 0.0
            delete_count = 0

            for i in range(num_nodes):
                key = i.to_bytes(4, byteorder="big")
                value = (2 * i).to_bytes(4, byteorder="big")
                seed = leaf_hash(key=key, value=value)
                reference_node_hash: Optional[bytes32] = await data_store.get_terminal_node_for_seed(store_id, seed)
                side: Optional[Side] = data_store.get_side_for_seed(seed)

                if i == 0:
                    reference_node_hash = None
                    side = None
                if i % 3 == 0:
                    t1 = time.time()
                    if not slow_mode:
                        await data_store.insert(
                            key=key,
                            value=value,
                            store_id=store_id,
                            reference_node_hash=reference_node_hash,
                            side=side,
                        )
                    else:
                        await data_store.insert(
                            key=key,
                            value=value,
                            store_id=store_id,
                            reference_node_hash=reference_node_hash,
                            side=side,
                            use_optimized=False,
                        )
                    t2 = time.time()
                    insert_time += t2 - t1
                    insert_count += 1
                elif i % 3 == 1:
                    t1 = time.time()
                    if not slow_mode:
                        await data_store.autoinsert(key=key, value=value, store_id=store_id)
                    else:
                        await data_store.autoinsert(
                            key=key,
                            value=value,
                            store_id=store_id,
                            use_optimized=False,
                        )
                    t2 = time.time()
                    autoinsert_time += t2 - t1
                    autoinsert_count += 1
                else:
                    t1 = time.time()
                    assert reference_node_hash is not None
                    node = await data_store.get_node(reference_node_hash)
                    assert isinstance(node, TerminalNode)
                    if not slow_mode:
                        await data_store.delete(key=node.key, store_id=store_id)
                    else:
                        await data_store.delete(key=node.key, store_id=store_id, use_optimized=False)
                    t2 = time.time()
                    delete_time += t2 - t1
                    delete_count += 1

            print(f"Average insert time: {insert_time / insert_count}")
            print(f"Average autoinsert time: {autoinsert_time / autoinsert_count}")
            print(f"Average delete time: {delete_time / delete_count}")
            print(f"Total time for {num_nodes} operations: {insert_time + autoinsert_time + delete_time}")
            root = await data_store.get_tree_root(store_id=store_id)
            print(f"Root hash: {root.node_hash}")


if __name__ == "__main__":
    slow_mode = False
    if len(sys.argv) > 2 and sys.argv[2] == "slow":
        slow_mode = True
    asyncio.run(generate_datastore(int(sys.argv[1]), slow_mode))



from __future__ import annotations

import asyncio
import itertools
import logging
import pathlib
import random
import sys
import time

from chik._tests.util.misc import create_logger
from chik.util.task_referencer import create_referenced_task

# TODO: CAMPid 0945094189459712842390t591
IP = "127.0.0.1"
PORT = 9678
NUM_CLIENTS = 500

total_open_connections = 0


async def tcp_echo_client(task_counter: str, logger: logging.Logger) -> None:
    global total_open_connections
    try:
        for loop_counter in itertools.count():
            label = f"{task_counter:5}-{loop_counter:5}"
            await asyncio.sleep(random.random())
            t1 = time.monotonic()
            writer = None
            try:
                logger.info(f"Opening connection: {label}")
                _reader, writer = await asyncio.open_connection(IP, PORT)
                total_open_connections += 1
                logger.info(f"Opened connection: {label} (total: {total_open_connections})")
                assert writer is not None
                await asyncio.sleep(1 + 4 * random.random())
            except asyncio.CancelledError as e:
                t2 = time.monotonic()
                logger.info(f"Cancelled connection: {label} - {e}. Time: {t2 - t1:.3f}")
                break
            except Exception as e:
                t2 = time.monotonic()
                logger.info(f"Closed connection: {label} - {e}. Time: {t2 - t1:.3f}")
            finally:
                logger.info(f"--- {label} a")
                if writer is not None:
                    total_open_connections -= 1
                    logger.info(f"--- {label}   B (total: {total_open_connections})")
                    writer.close()
                    await writer.wait_closed()
    finally:
        logger.info(f"--- {task_counter:5} task finishing")


async def main() -> None:
    shutdown_path = pathlib.Path(sys.argv[1])
    out_path = shutdown_path.with_suffix(".out")

    async def dun() -> None:
        # TODO: switch to event driven code
        while shutdown_path.exists():  # noqa: ASYNC110
            await asyncio.sleep(0.25)

        task.cancel()

    file_task = create_referenced_task(dun())

    with out_path.open(mode="w") as file:
        logger = create_logger(file=file)

        async def f() -> None:
            await asyncio.gather(*[tcp_echo_client(task_counter=f"{i}", logger=logger) for i in range(0, NUM_CLIENTS)])

        task = create_referenced_task(f())
        try:
            await task
        except asyncio.CancelledError:
            pass
        finally:
            logger.info("leaving flood")
            await file_task


asyncio.run(main())
